# noqa: D100
from __future__ import annotations

import numpy as np
import xarray as xr

from xclim.core.calendar import get_calendar
from xclim.core.missing import at_least_n_valid
from xclim.core.units import declare_units, rate2amount, to_agg_units
from xclim.indices.generic import threshold_count

from . import generic

__all__ = [
    "base_flow_index",
    "flow_index",
    "high_flow_frequency",
    "low_flow_frequency",
    "melt_and_precip_max",
    "rb_flashiness_index",
    "snd_max",
    "snd_max_doy",
    "snow_melt_we_max",
    "snw_max",
    "snw_max_doy",
]


@declare_units(q="[discharge]")
def base_flow_index(q: xr.DataArray, freq: str = "YS") -> xr.DataArray:
    r"""Base flow index.

    Return the base flow index, defined as the minimum 7-day average flow divided by the mean flow.

    Parameters
    ----------
    q : xarray.DataArray
        Rate of river discharge.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Base flow index.

    Notes
    -----
    Let :math:`\mathbf{q}=q_0, q_1, \ldots, q_n` be the sequence of daily discharge and :math:`\overline{\mathbf{q}}`
    the mean flow over the period. The base flow index is given by:

    .. math::

       \frac{\min(\mathrm{CMA}_7(\mathbf{q}))}{\overline{\mathbf{q}}}


    where :math:`\mathrm{CMA}_7` is the seven days moving average of the daily flow:

    .. math::

       \mathrm{CMA}_7(q_i) = \frac{\sum_{j=i-3}^{i+3} q_j}{7}

    """
    m7 = q.rolling(time=7, center=True).mean(skipna=False).resample(time=freq)
    mq = q.resample(time=freq)

    m7m = m7.min(dim="time")
    out = m7m / mq.mean(dim="time")
    out.attrs["units"] = ""
    return out


@declare_units(q="[discharge]")
def rb_flashiness_index(q: xr.DataArray, freq: str = "YS") -> xr.DataArray:
    r"""Richards-Baker flashiness index.

    Measures oscillations in flow relative to total flow, quantifying the frequency and rapidity of short term changes
    in flow, based on :cite:t:`baker_new_2004`.

    Parameters
    ----------
    q : xarray.DataArray
        Rate of river discharge.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        R-B Index.

    Notes
    -----
    Let :math:`\mathbf{q}=q_0, q_1, \ldots, q_n` be the sequence of daily discharge, the R-B Index is given by:

    .. math::

       \frac{\sum_{i=1}^n |q_i - q_{i-1}|}{\sum_{i=1}^n q_i}

    References
    ----------
    :cite:cts:`baker_new_2004`
    """
    d = np.abs(q.diff(dim="time")).resample(time=freq)
    mq = q.resample(time=freq)
    out = d.sum(dim="time") / mq.sum(dim="time")
    out.attrs["units"] = ""
    return out


@declare_units(snd="[length]")
def snd_max(snd: xr.DataArray, freq: str = "YS-JUL") -> xr.DataArray:
    """Maximum snow depth.

    The maximum daily snow depth.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow depth (mass per area).
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow depth over a given number of days for each period. [length].
    """
    return generic.select_resample_op(snd, op="max", freq=freq)


@declare_units(snd="[length]")
def snd_max_doy(snd: xr.DataArray, freq: str = "YS-JUL") -> xr.DataArray:
    """Maximum snow depth day of year.

    Day of year when surface snow reaches its peak value. If snow depth is 0 over entire period, return NaN.

    Parameters
    ----------
    snd : xarray.DataArray
        Surface snow depth.
    freq : str
         Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The day of year at which snow depth reaches its maximum value.
    """
    # Identify periods where there is at least one non-null value for snow depth
    valid = at_least_n_valid(snd.where(snd > 0), n=1, freq=freq)

    # Compute doymax. Will return first time step if all snow depths are 0.
    out = generic.select_resample_op(
        snd.where(snd > 0, 0), op=generic.doymax, freq=freq
    )
    out.attrs.update(units="", is_dayofyear=np.int32(1), calendar=get_calendar(snd))

    # Mask arrays that miss at least one non-null snd.
    return out.where(~valid)


@declare_units(snw="[mass]/[area]")
def snw_max(snw: xr.DataArray, freq: str = "YS-JUL") -> xr.DataArray:
    """Maximum snow amount.

    The maximum daily snow amount.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow amount (mass per area).
    freq: str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow amount over a given number of days for each period. [mass/area].
    """
    return generic.select_resample_op(snw, op="max", freq=freq)


@declare_units(snw="[mass]/[area]")
def snw_max_doy(snw: xr.DataArray, freq: str = "YS-JUL") -> xr.DataArray:
    """Maximum snow amount day of year.

    Day of year when surface snow amount reaches its peak value. If snow amount is 0 over entire period, return NaN.

    Parameters
    ----------
    snw : xarray.DataArray
        Surface snow amount.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The day of year at which snow amount reaches its maximum value.
    """
    # Identify periods where there is at least one non-null value for snow depth
    valid = at_least_n_valid(snw.where(snw > 0), n=1, freq=freq)

    # Compute doymax. Will return first time step if all snow depths are 0.
    out = generic.select_resample_op(
        snw.where(snw > 0, 0), op=generic.doymax, freq=freq
    )
    out.attrs.update(units="", is_dayofyear=np.int32(1), calendar=get_calendar(snw))

    # Mask arrays that miss at least one non-null snd.
    return out.where(~valid)


@declare_units(snw="[mass]/[area]")
def snow_melt_we_max(
    snw: xr.DataArray, window: int = 3, freq: str = "YS-JUL"
) -> xr.DataArray:
    """Maximum snow melt.

    The maximum snow melt over a given number of days expressed in snow water equivalent.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow amount (mass per area).
    window : int
        Number of days during which the melt is accumulated.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow melt over a given number of days for each period. [mass/area].
    """
    # Compute change in SWE. Set melt as a positive change.
    dsnw = snw.diff(dim="time") * -1

    # Sum over window
    agg = dsnw.rolling(time=window).sum()

    # Max over period
    out = agg.resample(time=freq).max(dim="time")
    out.attrs["units"] = snw.units
    return out


@declare_units(snw="[mass]/[area]", pr="[precipitation]")
def melt_and_precip_max(
    snw: xr.DataArray, pr: xr.DataArray, window: int = 3, freq: str = "YS-JUL"
) -> xr.DataArray:
    """Maximum snow melt and precipitation.

    The maximum snow melt plus precipitation over a given number of days expressed in snow water equivalent.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow amount (mass per area).
    pr : xarray.DataArray
        Daily precipitation flux.
    window : int
        Number of days during which the water input is accumulated.
    freq: str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow melt plus precipitation over a given number of days for each period. [mass/area].
    """
    # Compute change in SWE. Set melt as a positive change.
    dsnw = snw.diff(dim="time") * -1

    # Add precipitation total
    total = rate2amount(pr) + dsnw

    # Sum over window
    agg = total.rolling(time=window).sum()

    # Max over period
    out = agg.resample(time=freq).max(dim="time")
    out.attrs["units"] = snw.units
    return out


@declare_units(q="[discharge]")
def flow_index(q: xr.DataArray, p: float = 0.95) -> xr.DataArray:
    """
    Flow index

    Calculate the pth percentile of daily streamflow normalized by the median flow.

    Parameters
    ----------
    q : xr.DataArray
        Daily streamflow data.
    p : float
        Percentile for calculating the flow index, between 0 and 1. Default of 0.95 is for high flows.

    Returns
    -------
    xr.DataArray
        Normalized Qp, which is the p th percentile of daily streamflow normalized by the median flow.

    References
    ----------
    :cite:cts:`Clausen2000`
    """
    qp = q.quantile(p, dim="time")
    q_median = q.median(dim="time")
    out = qp / q_median
    out.attrs["units"] = "1"
    return out


@declare_units(q="[discharge]")
def high_flow_frequency(
    q: xr.DataArray, threshold_factor: int = 9, freq: str = "YS-OCT"
) -> xr.DataArray:
    """
    High flow frequency.

    Calculate the number of days in a given period with flows greater than a specified threshold, given as a
    multiple of the median flow. By default, the period is the water year starting on 1st October and ending on
    30th September, as commonly defined in North America.

    Parameters
    ----------
    q : xr.DataArray
        Daily streamflow data.
    threshold_factor : int
        Factor by which the median flow is multiplied to set the high flow threshold, default is 9.
    freq : str, optional
        Resampling frequency, default is 'YS-OCT' for water year starting in October and ending in September.

    Returns
    -------
    xr.DataArray
        Number of high flow days.

    References
    ----------
    :cite:cts:`addor2018,Clausen2000`
    """
    median_flow = q.median(dim="time")
    threshold = threshold_factor * median_flow
    out = threshold_count(q, ">", threshold, freq=freq)
    return to_agg_units(out, q, "count")


@declare_units(q="[discharge]")
def low_flow_frequency(
    q: xr.DataArray, threshold_factor: float = 0.2, freq: str = "YS-OCT"
) -> xr.DataArray:
    """
    Low flow frequency.

    Calculate the number of days in a given period with flows lower than a specified threshold, given by a fraction
    of the mean flow. By default, the period is the water year starting on 1st October and ending on 30th September,
    as commonly defined in North America.

    Parameters
    ----------
    q : xr.DataArray
        Daily streamflow data.
    threshold_factor : float
        Factor by which the mean flow is multiplied to set the low flow threshold, default is 0.2.
    freq : str, optional
        Resampling frequency, default is 'YS-OCT' for water year starting in October and ending in September.

    Returns
    -------
    xr.DataArray
        Number of low flow days.

    References
    ----------
    :cite:cts:`Olden2003`
    """
    mean_flow = q.mean(dim="time")
    threshold = threshold_factor * mean_flow
    out = threshold_count(q, "<", threshold, freq=freq)
    return to_agg_units(out, q, "count")



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from xarray import set_options

from xclim import sdba
from xclim.core.units import convert_units_to


class TestProperties:
    def test_mean(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1980"), location="Vancouver")
            .pr
        ).load()

        out_year = sdba.properties.mean(sim)
        np.testing.assert_array_almost_equal(out_year.values, [3.0016028e-05])

        out_season = sdba.properties.mean(sim, group="time.season")
        np.testing.assert_array_almost_equal(
            out_season.values,
            [4.6115547e-05, 1.7220482e-05, 2.8805329e-05, 2.825359e-05],
        )

        assert out_season.long_name.startswith("Mean")

    def test_var(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1980"), location="Vancouver")
            .pr
        ).load()

        out_year = sdba.properties.var(sim)
        np.testing.assert_array_almost_equal(out_year.values, [2.5884779e-09])

        out_season = sdba.properties.var(sim, group="time.season")
        np.testing.assert_array_almost_equal(
            out_season.values,
            [3.9270796e-09, 1.2538864e-09, 1.9057025e-09, 2.8776632e-09],
        )
        assert out_season.long_name.startswith("Variance")
        assert out_season.units == "kg2 m-4 s-2"

    def test_std(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1980"), location="Vancouver")
            .pr
        ).load()

        out_year = sdba.properties.std(sim)
        np.testing.assert_array_almost_equal(out_year.values, [5.08770208398345e-05])

        out_season = sdba.properties.std(sim, group="time.season")
        np.testing.assert_array_almost_equal(
            out_season.values,
            [6.2666411e-05, 3.5410259e-05, 4.3654352e-05, 5.3643853e-05],
        )
        assert out_season.long_name.startswith("Standard deviation")
        assert out_season.units == "kg m-2 s-1"

    def test_skewness(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1980"), location="Vancouver")
            .pr
        ).load()

        out_year = sdba.properties.skewness(sim)
        np.testing.assert_array_almost_equal(out_year.values, [2.8497460898513745])

        out_season = sdba.properties.skewness(sim, group="time.season")
        np.testing.assert_array_almost_equal(
            out_season.values,
            [
                2.036650744163691,
                3.7909534745807147,
                2.416590445325826,
                3.3521301798559566,
            ],
        )
        assert out_season.long_name.startswith("Skewness")
        assert out_season.units == ""

    def test_quantile(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1980"), location="Vancouver")
            .pr
        ).load()

        out_year = sdba.properties.quantile(sim, q=0.2)
        np.testing.assert_array_almost_equal(out_year.values, [2.8109431013945154e-07])

        out_season = sdba.properties.quantile(sim, group="time.season", q=0.2)
        np.testing.assert_array_almost_equal(
            out_season.values,
            [
                1.5171653330980917e-06,
                9.822543773907455e-08,
                1.8135805248675763e-07,
                4.135342521749408e-07,
            ],
        )
        assert out_season.long_name.startswith("Quantile 0.2")

    # TODO: test theshold_count? it's the same a test_spell_length_distribution
    def test_spell_length_distribution(self, open_dataset):
        ds = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .load()
        )

        # test pr, with amount method
        sim = ds.pr
        kws = {"op": "<", "group": "time.month"}
        outd = {
            stat: sdba.properties.spell_length_distribution(da=sim, **kws, stat=stat)
            .sel(month=1)
            .values
            for stat in ["mean", "max", "min"]
        }
        np.testing.assert_array_almost_equal(
            [outd[k] for k in ["mean", "max", "min"]], [2.44127, 10, 1]
        )

        # test tasmax, with quantile method
        simt = ds.tasmax
        kws = {"thresh": 0.9, "op": ">=", "method": "quantile", "group": "time.month"}
        outd = {
            stat: sdba.properties.spell_length_distribution(
                da=simt, **kws, stat=stat
            ).sel(month=6)
            for stat in ["mean", "max", "min"]
        }
        np.testing.assert_array_almost_equal(
            [outd[k].values for k in ["mean", "max", "min"]], [3.0, 6, 1]
        )

        # test varia
        with pytest.raises(
            ValueError,
            match="percentile is not a valid method. Choose 'amount' or 'quantile'.",
        ):
            sdba.properties.spell_length_distribution(simt, method="percentile")

        assert (
            outd["mean"].long_name
            == "Average of spell length distribution when the variable is >= the quantile 0.9 for 1 consecutive day(s)."
        )

    def test_spell_length_distribution_mixed_stat(self, open_dataset):

        time = pd.date_range("2000-01-01", periods=2 * 365, freq="D")
        tas = xr.DataArray(
            np.array([0] * 365 + [40] * 365),
            dims=("time"),
            coords={"time": time},
            attrs={"units": "degC"},
        )

        kws_sum = dict(
            thresh="30 degC", op=">=", stat="sum", stat_resample="sum", group="time"
        )
        out_sum = sdba.properties.spell_length_distribution(tas, **kws_sum).values
        kws_mixed = dict(
            thresh="30 degC", op=">=", stat="mean", stat_resample="sum", group="time"
        )
        out_mixed = sdba.properties.spell_length_distribution(tas, **kws_mixed).values

        assert out_sum == 365
        assert out_mixed == 182.5

    @pytest.mark.parametrize(
        "window,expected_amount,expected_quantile",
        [
            (1, [2.333333, 4, 1], [3, 6, 1]),
            (3, [1.333333, 4, 0], [2, 6, 0]),
        ],
    )
    def test_bivariate_spell_length_distribution(
        self, open_dataset, window, expected_amount, expected_quantile
    ):
        ds = (
            open_dataset("sdba/CanESM2_1950-2100.nc").sel(
                time=slice("1950", "1952"), location="Vancouver"
            )
        ).load()
        tx = ds.tasmax
        with set_options(keep_attrs=True):
            tn = tx - 5

        # test with amount method
        kws = {
            "thresh1": "0 degC",
            "thresh2": "0 degC",
            "op1": ">",
            "op2": "<=",
            "group": "time.month",
            "window": window,
        }
        outd = {
            stat: sdba.properties.bivariate_spell_length_distribution(
                da1=tx, da2=tn, **kws, stat=stat
            )
            .sel(month=1)
            .values
            for stat in ["mean", "max", "min"]
        }
        np.testing.assert_array_almost_equal(
            [outd[k] for k in ["mean", "max", "min"]], expected_amount
        )

        # test with quantile method
        kws = {
            "thresh1": 0.9,
            "thresh2": 0.9,
            "op1": ">",
            "op2": ">",
            "method1": "quantile",
            "method2": "quantile",
            "group": "time.month",
            "window": window,
        }
        outd = {
            stat: sdba.properties.bivariate_spell_length_distribution(
                da1=tx, da2=tn, **kws, stat=stat
            )
            .sel(month=6)
            .values
            for stat in ["mean", "max", "min"]
        }
        np.testing.assert_array_almost_equal(
            [outd[k] for k in ["mean", "max", "min"]], expected_quantile
        )

    def test_acf(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .pr
        ).load()

        out = sdba.properties.acf(sim, lag=1, group="time.month").sel(month=1)
        np.testing.assert_array_almost_equal(out.values, [0.11242357313756905])

        with pytest.raises(ValueError, match="Grouping period year is not allowed for"):
            sdba.properties.acf(sim, group="time")

        assert out.long_name.startswith("Lag-1 autocorrelation")
        assert out.units == ""

    def test_annual_cycle(self, open_dataset):
        simt = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .tasmax
        ).load()

        amp = sdba.properties.annual_cycle_amplitude(simt)
        relamp = sdba.properties.relative_annual_cycle_amplitude(simt)
        phase = sdba.properties.annual_cycle_phase(simt)

        np.testing.assert_allclose(
            [amp.values, relamp.values, phase.values],
            [16.74645996, 5.802083, 167],
            rtol=1e-5,
        )
        with pytest.raises(
            ValueError,
            match="Grouping period season is not allowed for property",
        ):
            sdba.properties.annual_cycle_amplitude(simt, group="time.season")

        with pytest.raises(
            ValueError,
            match="Grouping period month is not allowed for property",
        ):
            sdba.properties.annual_cycle_phase(simt, group="time.month")

        assert amp.long_name.startswith("Absolute amplitude of the annual cycle")
        assert phase.long_name.startswith("Phase of the annual cycle")
        assert amp.units == "K"
        assert amp.units_metadata == "temperature: difference"
        assert relamp.units == "%"
        assert phase.units == ""

    def test_annual_range(self, open_dataset):
        simt = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .tasmax
        ).load()

        # Initial annual cycle was this with window = 1
        amp = sdba.properties.mean_annual_range(simt, window=1)
        relamp = sdba.properties.mean_annual_relative_range(simt, window=1)
        phase = sdba.properties.mean_annual_phase(simt, window=1)

        np.testing.assert_allclose(
            [amp.values, relamp.values, phase.values],
            [34.039806, 11.793684020675501, 165.33333333333334],
        )

        amp = sdba.properties.mean_annual_range(simt)
        relamp = sdba.properties.mean_annual_relative_range(simt)
        phase = sdba.properties.mean_annual_phase(simt)

        np.testing.assert_array_almost_equal(
            [amp.values, relamp.values, phase.values],
            [18.715261, 6.480101, 181.6666667],
        )
        with pytest.raises(
            ValueError,
            match="Grouping period season is not allowed for property",
        ):
            sdba.properties.mean_annual_range(simt, group="time.season")

        with pytest.raises(
            ValueError,
            match="Grouping period month is not allowed for property",
        ):
            sdba.properties.mean_annual_phase(simt, group="time.month")

        assert amp.long_name.startswith("Average annual absolute amplitude")
        assert phase.long_name.startswith("Average annual phase")
        assert amp.units == "K"
        assert amp.units_metadata == "temperature: difference"
        assert relamp.units == "%"
        assert phase.units == ""

    def test_corr_btw_var(self, open_dataset):
        simt = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .tasmax
        ).load()

        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .pr
        ).load()

        pc = sdba.properties.corr_btw_var(simt, sim, corr_type="Pearson")
        pp = sdba.properties.corr_btw_var(
            simt, sim, corr_type="Pearson", output="pvalue"
        ).values
        sc = sdba.properties.corr_btw_var(simt, sim).values
        sp = sdba.properties.corr_btw_var(simt, sim, output="pvalue").values
        sc_jan = (
            sdba.properties.corr_btw_var(simt, sim, group="time.month")
            .sel(month=1)
            .values
        )
        sim[0] = np.nan
        pc_nan = sdba.properties.corr_btw_var(sim, simt, corr_type="Pearson").values

        np.testing.assert_array_almost_equal(
            [pc.values, pp, sc, sp, sc_jan, pc_nan],
            [
                -0.20849051347480407,
                3.2160438749049577e-12,
                -0.3449358561881698,
                5.97619379511559e-32,
                0.28329503745038936,
                -0.2090292,
            ],
        )
        assert pc.long_name == "Pearson correlation coefficient"
        assert pc.units == ""

        with pytest.raises(
            ValueError,
            match="pear is not a valid type. Choose 'Pearson' or 'Spearman'.",
        ):
            sdba.properties.corr_btw_var(sim, simt, group="time", corr_type="pear")

    def test_relative_frequency(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .pr
        ).load()

        test = sdba.properties.relative_frequency(sim, thresh="25 mm d-1", op=">=")
        testjan = (
            sdba.properties.relative_frequency(
                sim, thresh="25 mm d-1", op=">=", group="time.month"
            )
            .sel(month=1)
            .values
        )
        np.testing.assert_array_almost_equal(
            [test.values, testjan], [0.0045662100456621, 0.010752688172043012]
        )
        assert test.long_name == "Relative frequency of values >= 25 mm d-1."
        assert test.units == ""

    def test_transition(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .pr
        ).load()

        test = sdba.properties.transition_probability(
            da=sim, initial_op="<", final_op=">="
        )

        np.testing.assert_array_almost_equal([test.values], [0.14076782449725778])
        assert (
            test.long_name
            == "Transition probability of values < 1 mm d-1 to values >= 1 mm d-1."
        )
        assert test.units == ""

    def test_trend(self, open_dataset):
        simt = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "1952"), location="Vancouver")
            .tasmax
        ).load()

        slope = sdba.properties.trend(simt).values
        intercept = sdba.properties.trend(simt, output="intercept").values
        rvalue = sdba.properties.trend(simt, output="rvalue").values
        pvalue = sdba.properties.trend(simt, output="pvalue").values
        stderr = sdba.properties.trend(simt, output="stderr").values
        intercept_stderr = sdba.properties.trend(simt, output="intercept_stderr").values

        np.testing.assert_array_almost_equal(
            [slope, intercept, rvalue, pvalue, stderr, intercept_stderr],
            [
                -0.133711111111111,
                288.762132222222222,
                -0.9706433333333333,
                0.1546344444444444,
                0.033135555555555,
                0.042776666666666,
            ],
            4,
        )

        slope = sdba.properties.trend(simt, group="time.month").sel(month=1)
        intercept = (
            sdba.properties.trend(simt, output="intercept", group="time.month")
            .sel(month=1)
            .values
        )
        rvalue = (
            sdba.properties.trend(simt, output="rvalue", group="time.month")
            .sel(month=1)
            .values
        )
        pvalue = (
            sdba.properties.trend(simt, output="pvalue", group="time.month")
            .sel(month=1)
            .values
        )
        stderr = (
            sdba.properties.trend(simt, output="stderr", group="time.month")
            .sel(month=1)
            .values
        )
        intercept_stderr = (
            sdba.properties.trend(simt, output="intercept_stderr", group="time.month")
            .sel(month=1)
            .values
        )

        np.testing.assert_array_almost_equal(
            [slope.values, intercept, rvalue, pvalue, stderr, intercept_stderr],
            [
                0.8254511111111111,
                281.76353222222222,
                0.576843333333333,
                0.6085644444444444,
                1.1689105555555555,
                1.509056666666666,
            ],
            4,
        )

        assert slope.long_name.startswith("Slope of the interannual linear trend")
        assert slope.units == "K/year"

    def test_return_value(self, open_dataset):
        simt = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1950", "2010"), location="Vancouver")
            .tasmax
        ).load()

        out_y = sdba.properties.return_value(simt)

        out_djf = (
            sdba.properties.return_value(simt, op="min", group="time.season")
            .sel(season="DJF")
            .values
        )

        np.testing.assert_array_almost_equal(
            [out_y.values, out_djf], [313.154, 278.072], 3
        )
        assert out_y.long_name.startswith("20-year maximal return level")

    @pytest.mark.slow
    def test_spatial_correlogram(self, open_dataset):
        # This also tests sdba.utils._pairwise_spearman and sdba.nbutils._pairwise_haversine_and_bins
        # Test 1, does it work with 1D data?
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1981", "2010"))
            .tasmax
        ).load()

        out = sdba.properties.spatial_correlogram(sim, dims=["location"], bins=3)
        np.testing.assert_allclose(out, [-1, np.nan, 0], atol=1e-6)

        # Test 2, not very exhaustive, this is more of a detect-if-we-break-it test.
        sim = open_dataset("NRCANdaily/nrcan_canada_daily_tasmax_1990.nc").tasmax
        out = sdba.properties.spatial_correlogram(
            sim.isel(lon=slice(0, 50)), dims=["lon", "lat"], bins=20
        )
        np.testing.assert_allclose(
            out[:5],
            [0.95099902, 0.83028772, 0.66874473, 0.48893958, 0.30915054],
        )
        np.testing.assert_allclose(
            out.distance[:5],
            [26.543199, 67.716227, 108.889254, 150.062282, 191.23531],
            rtol=5e-07,
        )

    @pytest.mark.slow
    def test_decorrelation_length(self, open_dataset):
        sim = (
            open_dataset("NRCANdaily/nrcan_canada_daily_tasmax_1990.nc")
            .tasmax.isel(lon=slice(0, 5), lat=slice(0, 1))
            .load()
        )

        out = sdba.properties.decorrelation_length(
            sim, dims=["lat", "lon"], bins=10, radius=30
        )
        np.testing.assert_allclose(
            out[0],
            [4.5, 4.5, 4.5, 4.5, 10.5],
        )

    def test_get_measure(self, open_dataset):
        sim = (
            open_dataset("sdba/CanESM2_1950-2100.nc")
            .sel(time=slice("1981", "2010"), location="Vancouver")
            .pr
        ).load()

        ref = (
            open_dataset("sdba/ahccd_1950-2013.nc")
            .sel(time=slice("1981", "2010"), location="Vancouver")
            .pr
        ).load()

        sim = convert_units_to(sim, ref, context="hydro")
        sim_var = sdba.properties.var(sim)
        ref_var = sdba.properties.var(ref)

        meas = sdba.properties.var.get_measure()(sim_var, ref_var)
        np.testing.assert_allclose(meas, [0.408327], rtol=1e-3)



from __future__ import annotations

import datetime as dt
import re

from xclim import __version__
from xclim.core import formatting as fmt
from xclim.indicators.atmos import degree_days_exceedance_date, heat_wave_frequency


def test_prefix_attrs():
    source = {"units": "mm/s", "name": "pr"}
    dest = fmt.prefix_attrs(source, ["units"], "original_")
    assert "original_units" in dest

    out = fmt.unprefix_attrs(dest, ["units"], "original_")
    assert out == source

    # Check that the "naked" units will be overwritten.
    dest["units"] = ""

    out = fmt.unprefix_attrs(dest, ["units"], "original_")
    assert out == source


def test_indicator_docstring():
    doc = heat_wave_frequency.__doc__.split("\n")
    assert doc[0] == "Heat wave frequency (realm: atmos)"
    assert (
        doc[5]
        == "Based on indice :py:func:`~xclim.indices._multivariate.heat_wave_frequency`."
    )
    assert doc[6] == "Keywords : temperature health,."
    assert doc[12] == "    Default : `ds.tasmin`. [Required units : [temperature]]"
    assert (
        doc[41]
        == "    Total number of series of at least {window} consecutive days with daily minimum temperature above "
        "{thresh_tasmin} and daily maximum temperature above {thresh_tasmax}, "
        "with additional attributes: **description**: {freq} number of heat wave events within a given period. "
        "A heat wave occurs when daily minimum and maximum temperatures exceed {thresh_tasmin} and {thresh_tasmax}, "
        "respectively, over at least {window} days."
    )

    doc = degree_days_exceedance_date.__doc__.split("\n")
    assert doc[21] == "    Default : >. "


def test_update_xclim_history(atmosds):
    @fmt.update_xclim_history
    def func(da, arg1, arg2=None, arg3=None):  # noqa: F841
        return da

    out = func(atmosds.tas, 1, arg2=[1, 2], arg3=None)

    matches = re.match(
        r"\[([0-9-:\s]*)]\s(\w*):\s(\w*)\((.*)\)\s-\sxclim\sversion:\s(\d*\.\d*\.\d*[a-zA-Z-]*(\.\d*)?)",
        out.attrs["history"],
    ).groups()

    date = dt.datetime.fromisoformat(matches[0])
    assert dt.timedelta(0) < (dt.datetime.now() - date) < dt.timedelta(seconds=3)
    assert matches[1] == "tas"
    assert matches[2] == "func"
    assert matches[3] == "da=tas, arg1=1, arg2=[1, 2], arg3=None"
    assert matches[4] == __version__



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from xclim.core import missing

K2C = 273.15


class TestMissingBase:
    """The base class is well tested for daily input through the subclasses."""

    def test_3hourly_input(self, random):
        """Creating array with 21 days of 3h"""
        n = 21 * 8
        time = xr.cftime_range(start="2002-01-01", periods=n, freq="3h")
        ts = xr.DataArray(random.random(n), dims="time", coords={"time": time})
        mb = missing.MissingBase(ts, freq="MS", src_timestep="3h")
        # Make sure count is 31  * 8, because we're requesting a MS freq.
        assert mb.count == 31 * 8

    def test_monthly_input(self, random):
        """Creating array with 11 months."""
        n = 11
        time = xr.cftime_range(start="2002-01-01", periods=n, freq="ME")
        ts = xr.DataArray(random.random(n), dims="time", coords={"time": time})
        mb = missing.MissingBase(ts, freq="YS", src_timestep="ME")
        # Make sure count is 12, because we're requesting a YS freq.
        assert mb.count == 12

        n = 5
        time = xr.cftime_range(start="2002-06-01", periods=n, freq="MS")
        ts = xr.DataArray(random.random(n), dims="time", coords={"time": time})
        mb = missing.MissingBase(ts, freq="YS", src_timestep="MS", season="JJA")
        assert mb.count == 3

    def test_seasonal_input(self, random):
        """Creating array with 11 seasons."""
        n = 11
        time = xr.cftime_range(start="2002-04-01", periods=n, freq="QS-JAN")
        ts = xr.DataArray(random.random(n), dims="time", coords={"time": time})
        mb = missing.MissingBase(ts, freq="YS", src_timestep="QS-JAN")
        # Make sure count is 12, because we're requesting a YS freq.
        np.testing.assert_array_equal(mb.count, [4, 4, 4, 1])

        with pytest.raises(
            NotImplementedError,
            match="frequency that is not aligned with the source timestep.",
        ):
            missing.MissingBase(ts, freq="YS", src_timestep="QS-DEC")


class TestMissingAnyFills:
    def test_missing_days(self, tas_series):
        a = np.arange(360.0)
        a[5:10] = np.nan
        ts = tas_series(a)
        out = missing.missing_any(ts, freq="MS")
        assert out[0]
        assert not out[1]

    def test_missing_months(self):
        n = 66
        times = pd.date_range("2001-12-30", freq="1D", periods=n)
        da = xr.DataArray(np.arange(n), [("time", times)])
        miss = missing.missing_any(da, "MS")
        np.testing.assert_array_equal(miss, [True, False, False, True])

    def test_missing_years(self):
        n = 378
        times = pd.date_range("2001-12-31", freq="1D", periods=n)
        da = xr.DataArray(np.arange(n), [("time", times)])
        miss = missing.missing_any(da, "YS")
        np.testing.assert_array_equal(miss, [True, False, True])

    def test_missing_season(self):
        n = 378
        times = pd.date_range("2001-12-31", freq="1D", periods=n)
        da = xr.DataArray(np.arange(n), [("time", times)])
        miss = missing.missing_any(da, "QE-NOV")
        np.testing.assert_array_equal(miss, [True, False, False, False, True])

    def test_to_period_start(self, tasmin_series):
        a = np.zeros(365) + K2C + 5.0
        a[2] -= 20
        ts = tasmin_series(a)
        miss = missing.missing_any(ts, freq="YS-JUL")
        np.testing.assert_equal(miss, [False])

    def test_to_period_end(self, tasmin_series):
        a = np.zeros(365) + K2C + 5.0
        a[2] -= 20
        ts = tasmin_series(a)
        miss = missing.missing_any(ts, freq="YE-JUN")
        np.testing.assert_equal(miss, [False])

    def test_month(self, tasmin_series):
        ts = tasmin_series(np.zeros(36))
        miss = missing.missing_any(ts, freq="YS", month=7)
        np.testing.assert_equal(miss, [False])

        miss = missing.missing_any(ts, freq="YS", month=8)
        np.testing.assert_equal(miss, [True])

        with pytest.raises(ValueError, match=r"No data for selected period."):
            missing.missing_any(ts, freq="YS", month=1)

        miss = missing.missing_any(ts, freq="YS", month=[7, 8])
        np.testing.assert_equal(miss, [True])

        ts = tasmin_series(np.zeros(76))
        miss = missing.missing_any(ts, freq="YS", month=[7, 8])
        np.testing.assert_equal(miss, [False])

    @pytest.mark.parametrize("calendar", ("proleptic_gregorian", "noleap", "360_day"))
    def test_season(self, tasmin_series, calendar):
        ts = tasmin_series(np.zeros(360))
        ts = ts.convert_calendar(calendar, missing=0, align_on="date")

        miss = missing.missing_any(ts, freq="YS", season="MAM")
        np.testing.assert_equal(miss, [False])

        miss = missing.missing_any(ts, freq="YS", season="JJA")
        np.testing.assert_array_equal(miss, [True, True])

        miss = missing.missing_any(ts, freq="YS", season="SON")
        np.testing.assert_equal(miss, [False])

    def test_no_freq(self, tasmin_series):
        ts = tasmin_series(np.zeros(360))

        miss = missing.missing_any(ts, freq=None)
        np.testing.assert_array_equal(miss, False)

        t = list(range(31))
        t.pop(5)
        ts2 = ts.isel(time=t)
        miss = missing.missing_any(ts2, freq=None, src_timestep="h")
        np.testing.assert_array_equal(miss, True)

        # With indexer
        miss = missing.missing_any(ts, freq=None, month=[7])
        np.testing.assert_array_equal(miss, False)

        miss = missing.missing_any(ts2, freq=None, month=[7], src_timestep="h")
        np.testing.assert_array_equal(miss, True)

    def test_hydro(self, open_dataset):
        ds = open_dataset("Raven/q_sim.nc")
        miss = missing.missing_any(ds.q_sim, freq="YS")
        np.testing.assert_array_equal(miss[:-1], False)
        np.testing.assert_array_equal(miss[-1], True)

        miss = missing.missing_any(ds.q_sim, freq="YS", season="JJA")
        np.testing.assert_array_equal(miss, False)

    def test_hourly(self, pr_hr_series):
        a = np.arange(2.0 * 32 * 24)
        a[5:10] = np.nan
        pr = pr_hr_series(a)
        out = missing.missing_any(pr, freq="MS")
        np.testing.assert_array_equal(out, [True, False, True])

    def test_seasonal(self, random):
        n = 11
        time = xr.cftime_range(start="2002-01-01", periods=n, freq="QS-JAN")
        ts = xr.DataArray(random.random(n), dims="time", coords={"time": time})
        out = missing.missing_any(ts, freq="YS")
        np.testing.assert_array_equal(out, [False, False, True])


class TestMissingWMO:
    def test_missing_days(self, tas_series):
        a = np.arange(360.0)
        a[5:7] = np.nan  # Number of missing values under acceptable limit in a month
        a[40:45] = np.nan  # Too many consecutive missing values
        a[70:92:2] = np.nan  # Too many non-consecutive missing values
        ts = tas_series(a)
        out = missing.missing_wmo(ts, freq="MS")
        assert not out[0]
        assert out[1]
        assert out[2]

    def test_missing_days_in_quarter(self, tas_series):
        a = np.arange(360.0)
        a[5:16] = np.nan  # Number of missing values under acceptable limit in a month
        ts = tas_series(a)
        out = missing.missing_wmo(ts, freq="QS-JAN")
        np.testing.assert_array_equal(out, [True, False, False, True])

    def test_hourly(self, pr_hr_series):
        pr = pr_hr_series(np.zeros(30))

        with pytest.raises(ValueError):
            missing.missing_wmo(pr, freq="MS")

    def test_incomplete_year(self, tas_series):
        # One complete month
        a = np.arange(31)
        ts = tas_series(a)
        out = missing.missing_wmo(ts, freq="YS")
        np.testing.assert_array_equal(out, [True])


class TestMissingPct:
    def test_missing_days(self, tas_series):
        a = np.arange(360.0)
        a[5:7] = np.nan  # Number of missing values under acceptable limit in a month
        a[40:45] = np.nan  # Too many missing values
        ts = tas_series(a)
        out = missing.missing_pct(ts, freq="MS", tolerance=0.1)
        assert not out[0]
        assert out[1]

    def test_hourly(self, pr_hr_series):
        a = np.arange(2.0 * 32 * 24)
        a[5:50] = np.nan
        pr = pr_hr_series(a)
        out = missing.missing_pct(pr, freq="MS", tolerance=0.01)
        np.testing.assert_array_equal(out, [True, False, True])

    def test_missing_period(self, tas_series):
        tas = tas_series(np.ones(366), start="2000-01-01")
        tas = tas.sel(time=tas.time.dt.month.isin([1, 2, 3, 4, 12]))
        out = missing.missing_pct(tas, freq="MS", tolerance=0.9, src_timestep="D")
        np.testing.assert_array_equal(out, [False] * 4 + [True] * 7 + [False])


class TestAtLeastNValid:
    def test_at_least_n_valid(self, tas_series):
        a = np.arange(360.0)
        a[5:10] = np.nan  # Number of missing values under acceptable limit in a month
        a[40:55] = np.nan  # Too many missing values
        ts = tas_series(a)
        out = missing.at_least_n_valid(ts, freq="MS", n=20)
        np.testing.assert_array_equal(out[:2], [False, True])

    def test_hourly(self, pr_hr_series):
        a = np.arange(2.0 * 32 * 24)
        a[0:240] = np.nan
        pr = pr_hr_series(a)
        out = missing.at_least_n_valid(pr, freq="MS", n=25 * 24)
        np.testing.assert_array_equal(out, [True, False, True])

    def test_missing_period(self, tas_series):
        tas = tas_series(np.ones(366), start="2000-01-01")
        tas = tas.sel(time=tas.time.dt.month.isin([1, 2, 3, 4, 12]))
        out = missing.missing_pct(tas, freq="MS", tolerance=0.9, src_timestep="D")
        np.testing.assert_array_equal(out, [False] * 4 + [True] * 7 + [False])


class TestHourly:
    """Test that missing algorithms also work on resampling from hourly to daily."""

    def pr(self, pr_hr_series):
        a = np.arange(24.0 * 10)
        a[10] = np.nan
        a[-12:] = np.nan
        return pr_hr_series(a)

    def test_any(self, pr_hr_series):
        pr = self.pr(pr_hr_series)
        out = missing.missing_any(pr, "D", src_timestep="h")
        np.testing.assert_array_equal(
            out,
            [True] + 8 * [False] + [True],
        )

    def test_pct(self, pr_hr_series):
        pr = self.pr(pr_hr_series)
        out = missing.missing_pct(pr, "D", src_timestep="h", tolerance=0.1)
        np.testing.assert_array_equal(
            out,
            9 * [False] + [True],
        )

    def test_at_least_n_valid(self, pr_hr_series):
        pr = self.pr(pr_hr_series)
        out = missing.at_least_n_valid(pr, "D", src_timestep="h", n=20)
        np.testing.assert_array_equal(
            out,
            9 * [False] + [True],
        )



"""
Formatting Utilities for Indicators
===================================
"""

from __future__ import annotations

import datetime as dt
import itertools
import re
import string
import warnings
from ast import literal_eval
from collections.abc import Callable, Sequence
from fnmatch import fnmatch
from inspect import _empty, signature  # noqa
from typing import Any

import xarray as xr
from boltons.funcutils import wraps

from xclim.core.utils import InputKind

DEFAULT_FORMAT_PARAMS = {
    "tasmin_per_thresh": "{unknown}",
    "tasmin_per_window": "{unknown}",
    "tasmin_per_period": "{unknown}",
    "tas_per_thresh": "{unknown}",
    "tas_per_window": "{unknown}",
    "tas_per_period": "{unknown}",
    "tasmax_per_thresh": "{unknown}",
    "tasmax_per_window": "{unknown}",
    "tasmax_per_period": "{unknown}",
    "pr_per_thresh": "{unknown}",
    "pr_per_window": "{unknown}",
    "pr_per_period": "{unknown}",
}


class AttrFormatter(string.Formatter):
    """A formatter for frequently used attribute values.

    See the doc of format_field() for more details.
    """

    def __init__(
        self,
        mapping: dict[str, Sequence[str]],
        modifiers: Sequence[str],
    ) -> None:
        """Initialize the formatter.

        Parameters
        ----------
        mapping : dict[str, Sequence[str]]
            A mapping from values to their possible variations.
        modifiers : Sequence[str]
            The list of modifiers, must be the as long as the longest value of `mapping`.
            Cannot include reserved modifier 'r'.
        """
        super().__init__()
        if "r" in modifiers:
            raise ValueError("Modifier 'r' is reserved for default raw formatting.")
        self.modifiers = modifiers
        self.mapping = mapping

    def format(self, format_string: str, /, *args: Any, **kwargs: Any) -> str:
        r"""Format a string.

        Parameters
        ----------
        format_string: str
        \*args: Any
        \*\*kwargs: Any

        Returns
        -------
        str
        """
        for k, v in DEFAULT_FORMAT_PARAMS.items():
            if k not in kwargs:
                kwargs.update({k: v})
        return super().format(format_string, *args, **kwargs)

    def format_field(self, value, format_spec):
        """Format a value given a formatting spec.

        If `format_spec` is in this Formatter's modifiers, the corresponding variation
        of value is given. If `format_spec` is 'r' (raw), the value is returned unmodified.
        If `format_spec` is not specified but `value` is in the mapping, the first variation is returned.

        Examples
        --------
        Let's say the string "The dog is {adj1}, the goose is {adj2}" is to be translated
        to French and that we know that possible values of `adj` are `nice` and `evil`.
        In French, the genre of the noun changes the adjective (cat = chat is masculine,
        and goose = oie is feminine) so we initialize the formatter as:

        >>> fmt = AttrFormatter(
        ...     {
        ...         "nice": ["beau", "belle"],
        ...         "evil": ["méchant", "méchante"],
        ...         "smart": ["intelligent", "intelligente"],
        ...     },
        ...     ["m", "f"],
        ... )
        >>> fmt.format(
        ...     "Le chien est {adj1:m}, l'oie est {adj2:f}, le gecko est {adj3:r}",
        ...     adj1="nice",
        ...     adj2="evil",
        ...     adj3="smart",
        ... )
        "Le chien est beau, l'oie est méchante, le gecko est smart"

        The base values may be given using unix shell-like patterns:

        >>> fmt = AttrFormatter(
        ...     {"YS-*": ["annuel", "annuelle"], "MS": ["mensuel", "mensuelle"]},
        ...     ["m", "f"],
        ... )
        >>> fmt.format(
        ...     "La moyenne {freq:f} est faite sur un échantillon {src_timestep:m}",
        ...     freq="YS-JUL",
        ...     src_timestep="MS",
        ... )
        'La moyenne annuelle est faite sur un échantillon mensuel'
        """
        baseval = self._match_value(value)
        if baseval is None:  # Not something we know how to translate
            if format_spec in self.modifiers + [
                "r"
            ]:  # Woops, however a known format spec was asked
                warnings.warn(
                    f"Requested formatting `{format_spec}` for unknown string `{value}`."
                )
                format_spec = ""
            return super().format_field(value, format_spec)
        # Thus, known value

        if not format_spec:  # (None or '') No modifiers, return first
            return self.mapping[baseval][0]

        if format_spec == "r":  # Raw modifier
            return super().format_field(value, "")

        if format_spec in self.modifiers:  # Known modifier
            if len(self.mapping[baseval]) == 1:  # But unmodifiable entry
                return self.mapping[baseval][0]
            # Known modifier, modifiable entry
            return self.mapping[baseval][self.modifiers.index(format_spec)]
        # Known value but unknown modifier, must be a built-in one, only works for the default val...
        return super().format_field(self.mapping[baseval][0], format_spec)

    def _match_value(self, value):
        if isinstance(value, str):
            for mapval in self.mapping.keys():
                if fnmatch(value, mapval):
                    return mapval
        return None


# Tag mappings between keyword arguments and long-form text.
default_formatter = AttrFormatter(
    {
        # Arguments to "freq"
        "D": ["daily", "days"],
        "YS": ["annual", "years"],
        "YS-*": ["annual", "years"],
        "MS": ["monthly", "months"],
        "QS-*": ["seasonal", "seasons"],
        # Arguments to "indexer"
        "DJF": ["winter"],
        "MAM": ["spring"],
        "JJA": ["summer"],
        "SON": ["fall"],
        "norm": ["Normal"],
        "m1": ["january"],
        "m2": ["february"],
        "m3": ["march"],
        "m4": ["april"],
        "m5": ["may"],
        "m6": ["june"],
        "m7": ["july"],
        "m8": ["august"],
        "m9": ["september"],
        "m10": ["october"],
        "m11": ["november"],
        "m12": ["december"],
        # Arguments to "op / reducer / stat" (for example for generic.stats)
        "integral": ["integrated", "integral"],
        "count": ["count"],
        "doymin": ["day of minimum"],
        "doymax": ["day of maximum"],
        "mean": ["average"],
        "max": ["maximal", "maximum"],
        "min": ["minimal", "minimum"],
        "sum": ["total", "sum"],
        "std": ["standard deviation"],
        "var": ["variance"],
        "absamp": ["absolute amplitude"],
        "relamp": ["relative amplitude"],
        # For when we are formatting indicator classes with empty options
        "<class 'inspect._empty'>": ["<empty>"],
    },
    ["adj", "noun"],
)


def parse_doc(doc: str) -> dict[str, str]:
    """Crude regex parsing reading an indice docstring and extracting information needed in indicator construction.

    The appropriate docstring syntax is detailed in :ref:`notebooks/extendxclim:Defining new indices`.

    Parameters
    ----------
    doc : str
        The docstring of an indice function.

    Returns
    -------
    dict
        A dictionary with all parsed sections.
    """
    if doc is None:
        return {}

    out = {}

    sections = re.split(r"(\w+\s?\w+)\n\s+-{3,50}", doc)  # obj.__doc__.split('\n\n')
    intro = sections.pop(0)
    if intro:
        intro_content = list(map(str.strip, intro.strip().split("\n\n")))
        if len(intro_content) == 1:
            out["title"] = intro_content[0]
        elif len(intro_content) >= 2:
            out["title"], abstract = intro_content[:2]
            out["abstract"] = " ".join(map(str.strip, abstract.splitlines()))

    for i in range(0, len(sections), 2):
        header, content = sections[i : i + 2]

        if header in ["Notes", "References"]:
            out[header.lower()] = content.replace("\n    ", "\n").strip()
        elif header == "Parameters":
            out["parameters"] = _parse_parameters(content)
        elif header == "Returns":
            rets = _parse_returns(content)
            if rets:
                meta = list(rets.values())[0]
                if "long_name" in meta:
                    out["long_name"] = meta["long_name"]
    return out


def _parse_parameters(section):
    """Parse the 'parameters' section of a docstring into a dictionary.

    Works by mapping the parameter name to its description and, potentially, to its set of choices.
    The type annotation are not parsed, except for fixed sets of values (listed as "{'a', 'b', 'c'}").
    The annotation parsing only accepts strings, numbers, `None` and `nan` (to represent `numpy.nan`).
    """
    curr_key = None
    params = {}
    for line in section.split("\n"):
        if line.startswith(" " * 6):  # description
            s = " " if params[curr_key]["description"] else ""
            params[curr_key]["description"] += s + line.strip()
        elif line.startswith(" " * 4) and ":" in line:  # param title
            name, annot = line.split(":", maxsplit=1)
            curr_key = name.strip()
            params[curr_key] = {"description": ""}
            match = re.search(r".*(\{.*\}).*", annot)
            if match:
                try:
                    choices = literal_eval(match.groups()[0])
                    params[curr_key]["choices"] = choices
                except ValueError:  # noqa: S110
                    # If the literal_eval fails, we just ignore the choices.
                    pass
    return params


def _parse_returns(section):
    """Parse the returns section of a docstring into a dictionary mapping the parameter name to its description."""
    curr_key = None
    params = {}
    for line in section.split("\n"):
        if line.strip():
            if line.startswith(" " * 6):  # long_name
                s = " " if params[curr_key]["long_name"] else ""
                params[curr_key]["long_name"] += s + line.strip()
            elif line.startswith(" " * 4):  # param title
                annot, *name = reversed(line.split(":", maxsplit=1))
                if name:
                    curr_key = name[0].strip()
                else:
                    curr_key = None
                params[curr_key] = {"long_name": ""}
                annot, *unit = annot.split(",", maxsplit=1)
                if unit:
                    params[curr_key]["units"] = unit[0].strip()
    return params


def merge_attributes(
    attribute: str,
    *inputs_list: xr.DataArray | xr.Dataset,
    new_line: str = "\n",
    missing_str: str | None = None,
    **inputs_kws: xr.DataArray | xr.Dataset,
) -> str:
    r"""Merge attributes from several DataArrays or Datasets.

    If more than one input is given, its name (if available) is prepended as: "<input name> : <input attribute>".

    Parameters
    ----------
    attribute : str
        The attribute to merge.
    inputs_list : xr.DataArray or xr.Dataset
        The datasets or variables that were used to produce the new object.
        Inputs given that way will be prefixed by their `name` attribute if available.
    new_line : str
        The character to put between each instance of the attributes. Usually, in CF-conventions,
        the history attributes uses '\\n' while cell_methods uses ' '.
    missing_str : str
        A string that is printed if an input doesn't have the attribute. Defaults to None, in which
        case the input is simply skipped.
    \*\*inputs_kws : xr.DataArray or xr.Dataset
        Mapping from names to the datasets or variables that were used to produce the new object.
        Inputs given that way will be prefixes by the passed name.

    Returns
    -------
    str
        The new attribute made from the combination of the ones from all the inputs.
    """
    inputs = []
    for in_ds in inputs_list:
        inputs.append((getattr(in_ds, "name", None), in_ds))
    inputs += list(inputs_kws.items())

    merged_attr = ""
    for in_name, in_ds in inputs:
        if attribute in in_ds.attrs or missing_str is not None:
            if in_name is not None and len(inputs) > 1:
                merged_attr += f"{in_name}: "
            merged_attr += in_ds.attrs.get(
                attribute, "" if in_name is None else missing_str
            )
            merged_attr += new_line

    if len(new_line) > 0:
        return merged_attr[: -len(new_line)]  # Remove the last added new_line
    return merged_attr


def update_history(
    hist_str: str,
    *inputs_list: xr.DataArray | xr.Dataset,
    new_name: str | None = None,
    **inputs_kws: xr.DataArray | xr.Dataset,
) -> str:
    r"""Return a history string with the timestamped message and the combination of the history of all inputs.

    The new history entry is formatted as "[<timestamp>] <new_name>: <hist_str> - xclim version: <xclim.__version__>."

    Parameters
    ----------
    hist_str : str
        The string describing what has been done on the data.
    \*inputs_list : xr.DataArray or xr.Dataset
        The datasets or variables that were used to produce the new object.
        Inputs given that way will be prefixed by their "name" attribute if available.
    new_name : str, optional
        The name of the newly created variable or dataset to prefix hist_msg.
    \*\*inputs_kws : xr.DataArray or xr.Dataset
        Mapping from names to the datasets or variables that were used to produce the new object.
        Inputs given that way will be prefixes by the passed name.

    Returns
    -------
    str
        The combine history of all inputs starting with `hist_str`.

    See Also
    --------
    merge_attributes
    """
    from xclim import (  # pylint: disable=cyclic-import,import-outside-toplevel
        __version__,
    )

    merged_history = merge_attributes(
        "history",
        *inputs_list,
        new_line="\n",
        missing_str="",
        **inputs_kws,
    )
    if len(merged_history) > 0 and not merged_history.endswith("\n"):
        merged_history += "\n"
    merged_history += (
        f"[{dt.datetime.now():%Y-%m-%d %H:%M:%S}] {new_name or ''}: "
        f"{hist_str} - xclim version: {__version__}"
    )
    return merged_history


def update_xclim_history(func: Callable):
    """Decorator that auto-generates and fills the history attribute.

    The history is generated from the signature of the function and added to the first output.
    Because of a limitation of the `boltons` wrapper, all arguments passed to the wrapped function
    will be printed as keyword arguments.
    """

    @wraps(func)
    def _call_and_add_history(*args, **kwargs):
        """Call the function and then generate and add the history attr."""
        outs = func(*args, **kwargs)

        if isinstance(outs, tuple):
            out = outs[0]
        else:
            out = outs

        if not isinstance(out, xr.DataArray | xr.Dataset):
            raise TypeError(
                f"Decorated `update_xclim_history` received a non-xarray output from {func.__name__}."
            )

        da_list = [arg for arg in args if isinstance(arg, xr.DataArray)]
        da_dict = {
            name: arg for name, arg in kwargs.items() if isinstance(arg, xr.DataArray)
        }

        # The wrapper hides how the user passed the arguments (positional or keyword)
        # Instead of having it all position, we have it all keyword-like for explicitness.
        bound_args = signature(func).bind(*args, **kwargs)
        attr = update_history(
            gen_call_string(func.__name__, **bound_args.arguments),
            *da_list,
            new_name=out.name if isinstance(out, xr.DataArray) else None,
            **da_dict,
        )
        out.attrs["history"] = attr
        return outs

    return _call_and_add_history


def gen_call_string(funcname: str, *args, **kwargs) -> str:
    r"""Generate a signature string for use in the history attribute.

    DataArrays and Dataset are replaced with their name, while Nones, floats, ints and strings are printed directly.
    All other objects have their type printed between < >.

    Arguments given through positional arguments are printed positionnally and those
    given through keywords are printed prefixed by their name.

    Parameters
    ----------
    funcname : str
        Name of the function
    \*args, \*\*kwargs
        Arguments given to the function.

    Example
    -------
    >>> A = xr.DataArray([1], dims=("x",), name="A")
    >>> gen_call_string("func", A, b=2.0, c="3", d=[10] * 100)
    "func(A, b=2.0, c='3', d=<list>)"
    """
    elements = []
    chain = itertools.chain(zip([None] * len(args), args, strict=False), kwargs.items())
    for name, val in chain:
        if isinstance(val, xr.DataArray):
            rep = val.name or "<array>"
        elif isinstance(val, int | float | str | bool) or val is None:
            rep = repr(val)
        else:
            rep = repr(val)
            if len(rep) > 50:
                rep = f"<{type(val).__name__}>"

        if name is not None:
            rep = f"{name}={rep}"

        elements.append(rep)

    return f"{funcname}({', '.join(elements)})"


def prefix_attrs(source: dict, keys: Sequence, prefix: str) -> dict:
    """Rename some keys of a dictionary by adding a prefix.

    Parameters
    ----------
    source : dict
        Source dictionary, for example data attributes.
    keys : sequence
        Names of keys to prefix.
    prefix : str
        Prefix to prepend to keys.

    Returns
    -------
    dict
        Dictionary of attributes with some keys prefixed.
    """
    out = {}
    for key, val in source.items():
        if key in keys:
            out[f"{prefix}{key}"] = val
        else:
            out[key] = val
    return out


def unprefix_attrs(source: dict, keys: Sequence, prefix: str) -> dict:
    """Remove prefix from keys in a dictionary.

    Parameters
    ----------
    source : dict
        Source dictionary, for example data attributes.
    keys : sequence
        Names of original keys for which prefix should be removed.
    prefix : str
        Prefix to remove from keys.

    Returns
    -------
    dict
        Dictionary of attributes whose keys were prefixed, with prefix removed.
    """
    out = {}
    n = len(prefix)
    for key, val in source.items():
        k = key[n:]
        if (k in keys) and key.startswith(prefix):
            out[k] = val
        elif key not in out:
            out[key] = val
    return out


KIND_ANNOTATION = {
    InputKind.VARIABLE: "str or DataArray",
    InputKind.OPTIONAL_VARIABLE: "str or DataArray, optional",
    InputKind.QUANTIFIED: "quantity (string or DataArray, with units)",
    InputKind.FREQ_STR: "offset alias (string)",
    InputKind.NUMBER: "number",
    InputKind.NUMBER_SEQUENCE: "number or sequence of numbers",
    InputKind.STRING: "str",
    InputKind.DAY_OF_YEAR: "date (string, MM-DD)",
    InputKind.DATE: "date (string, YYYY-MM-DD)",
    InputKind.BOOL: "boolean",
    InputKind.DICT: "dict",
    InputKind.DATASET: "Dataset, optional",
    InputKind.KWARGS: "",
    InputKind.OTHER_PARAMETER: "Any",
}


def _gen_parameters_section(
    parameters: dict[str, dict[str, Any]], allowed_periods: list[str] | None = None
) -> str:
    """Generate the "parameters" section of the indicator docstring.

    Parameters
    ----------
    parameters : dict
        Parameters dictionary (`Ind.parameters`).
    allowed_periods : list of str, optional
        Restrict parameters to specific periods. Default: None.

    Returns
    -------
    str
    """
    section = "Parameters\n----------\n"
    for name, param in parameters.items():
        desc_str = param.description
        if param.kind == InputKind.FREQ_STR:
            desc_str += (
                " See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset"
                "-aliases for available options."
            )
            if allowed_periods is not None:
                desc_str += (
                    f" Restricted to frequencies equivalent to one of {allowed_periods}"
                )
        if param.kind == InputKind.VARIABLE:
            defstr = f"Default : `ds.{param.default}`. "
        elif param.kind == InputKind.OPTIONAL_VARIABLE:
            defstr = ""
        elif param.default is not _empty:
            defstr = f"Default : {param.default}. "
        else:
            defstr = "Required. "
        if "choices" in param:
            annotstr = str(param.choices)
        else:
            annotstr = KIND_ANNOTATION[param.kind]
        if "units" in param and param.units is not None:
            unitstr = f"[Required units : {param.units}]"
        else:
            unitstr = ""
        section += f"{name} {': ' if annotstr else ''}{annotstr}\n    {desc_str}\n    {defstr}{unitstr}\n"
    return section


def _gen_returns_section(cf_attrs: Sequence[dict[str, Any]]) -> str:
    """Generate the "Returns" section of an indicator's docstring.

    Parameters
    ----------
    cf_attrs : Sequence[Dict[str, Any]]
        The list of attributes, usually Indicator.cf_attrs.

    Returns
    -------
    str
    """
    section = "Returns\n-------\n"
    for attrs in cf_attrs:
        if not section.endswith("\n"):
            section += "\n"
        section += f"{attrs['var_name']} : DataArray\n"
        section += f"    {attrs.get('long_name', '')}"
        if "standard_name" in attrs:
            section += f" ({attrs['standard_name']})"
        if "units" in attrs:
            section += f" [{attrs['units']}]"
        added_section = ""
        for key, attr in attrs.items():
            if key not in ["long_name", "standard_name", "units", "var_name"]:
                if callable(attr):
                    attr = "<Dynamically generated string>"
                added_section += f" **{key}**: {attr};"
        if added_section:
            section = f"{section}, with additional attributes:{added_section[:-1]}"
    section += "\n"
    return section


def generate_indicator_docstring(ind) -> str:
    """Generate an indicator's docstring from keywords.

    Parameters
    ----------
    ind
        Indicator instance

    Returns
    -------
    str
    """
    header = f"{ind.title} (realm: {ind.realm})\n\n{ind.abstract}\n"

    special = ""

    if hasattr(ind, "missing"):  # Only ResamplingIndicators
        special += f'This indicator will check for missing values according to the method "{ind.missing}".\n'
    if hasattr(ind.compute, "__module__"):
        special += f"Based on indice :py:func:`~{ind.compute.__module__}.{ind.compute.__name__}`.\n"
        if ind.injected_parameters:
            special += "With injected parameters: "
            special += ", ".join(
                [f"{k}={v}" for k, v in ind.injected_parameters.items()]
            )
            special += ".\n"
    if ind.keywords:
        special += f"Keywords : {ind.keywords}.\n"

    parameters = _gen_parameters_section(
        ind.parameters, getattr(ind, "allowed_periods", None)
    )

    returns = _gen_returns_section(ind.cf_attrs)

    extras = ""
    for section in ["notes", "references"]:
        if getattr(ind, section):
            extras += f"{section.capitalize()}\n{'-' * len(section)}\n{getattr(ind, section)}\n\n"

    doc = f"{header}\n{special}\n{parameters}\n{returns}\n{extras}"
    return doc


def get_percentile_metadata(data: xr.DataArray, prefix: str) -> dict[str, str]:
    """Get the metadata related to percentiles from the given DataArray as a dictionary.

    Parameters
    ----------
    data : xr.DataArray
        Must be a percentile DataArray, this means the necessary metadata
        must be available in its attributes and coordinates.
    prefix : str
        The prefix to be used in the metadata key.
        Usually this takes the form of "tasmin_per" or equivalent.

    Returns
    -------
    dict
        A mapping of the configuration used to compute these percentiles.
    """
    # handle case where da was created with `quantile()` method
    if "quantile" in data.coords:
        percs = data.coords["quantile"].values * 100
    elif "percentiles" in data.coords:
        percs = data.coords["percentiles"].values
    else:
        percs = "<unknown percentiles>"
    clim_bounds = data.attrs.get("climatology_bounds", "<unknown bounds>")

    return {
        f"{prefix}_thresh": percs,
        f"{prefix}_window": data.attrs.get("window", "<unknown window>"),
        f"{prefix}_period": clim_bounds,
    }



"""
==============
Ensemble tools
==============

This submodule defines some useful methods for dealing with ensembles of climate simulations.
In xclim, an "ensemble" is a `Dataset` or a `DataArray` where multiple climate realizations
or models are concatenated along the `realization` dimension.
"""

from __future__ import annotations

from xclim.ensembles._base import (
    create_ensemble,
    ensemble_mean_std_max_min,
    ensemble_percentiles,
)
from xclim.ensembles._partitioning import (
    fractional_uncertainty,
    hawkins_sutton,
    lafferty_sriver,
)
from xclim.ensembles._reduce import (
    kkz_reduce_ensemble,
    kmeans_reduce_ensemble,
    make_criteria,
    plot_rsqprofile,
)
from xclim.ensembles._robustness import (
    robustness_categories,
    robustness_coefficient,
    robustness_fractions,
)



# noqa: D100
from __future__ import annotations

from collections.abc import Callable
from typing import cast

import numpy as np
import xarray

from xclim.core import Quantified
from xclim.core.bootstrapping import percentile_bootstrap
from xclim.core.calendar import resample_doy
from xclim.core.units import (
    convert_units_to,
    declare_units,
    pint2cfattrs,
    rate2amount,
    str2pint,
    to_agg_units,
)
from xclim.indices import run_length as rl
from xclim.indices._conversion import rain_approximation, snowfall_approximation
from xclim.indices.generic import compare, select_resample_op, threshold_count

# Frequencies : YS: year start, QS-DEC: seasons starting in december, MS: month start
# See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

# -------------------------------------------------- #
# ATTENTION: ASSUME ALL INDICES WRONG UNTIL TESTED ! #
# -------------------------------------------------- #

__all__ = [
    "blowing_snow",
    "cold_and_dry_days",
    "cold_and_wet_days",
    "cold_spell_duration_index",
    "daily_temperature_range",
    "daily_temperature_range_variability",
    "days_over_precip_thresh",
    "extreme_temperature_range",
    "fraction_over_precip_thresh",
    "heat_wave_frequency",
    "heat_wave_max_length",
    "heat_wave_total_length",
    "high_precip_low_temp",
    "liquid_precip_ratio",
    "multiday_temperature_swing",
    "precip_accumulation",
    "precip_average",
    "rain_on_frozen_ground_days",
    "tg10p",
    "tg90p",
    "tn10p",
    "tn90p",
    "tx10p",
    "tx90p",
    "tx_tn_days_above",
    "warm_and_dry_days",
    "warm_and_wet_days",
    "warm_spell_duration_index",
    "water_cycle_intensity",
    "winter_rain_ratio",
]


@declare_units(tasmin="[temperature]", tasmin_per="[temperature]")
@percentile_bootstrap
def cold_spell_duration_index(
    tasmin: xarray.DataArray,
    tasmin_per: xarray.DataArray,
    window: int = 6,
    freq: str = "YS",
    resample_before_rl: bool = True,
    bootstrap: bool = False,  # noqa  # noqa
    op: str = "<",
) -> xarray.DataArray:
    r"""Cold spell duration index.

    Number of days with at least `window` consecutive days when the daily minimum temperature is below the
    `tasmin_per` percentiles.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmin_per : xarray.DataArray
        nth percentile of daily minimum temperature with `dayofyear` coordinate.
    window : int
        Minimum number of days with temperature below threshold to qualify as a cold spell.
    freq : str
        Resampling frequency.
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {"<", "<=", "lt", "le"}
        Comparison operation. Default: "<".

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with at least six consecutive days when the daily minimum temperature is below the 10th
        percentile.

    Notes
    -----
    Let :math:`TN_i` be the minimum daily temperature for the day of the year :math:`i` and :math:`TN10_i` the 10th
    percentile of the minimum daily temperature over the 1961-1990 period for day of the year :math:`i`, the cold spell
    duration index over period :math:`\phi` is defined as:

    .. math::

       \sum_{i \in \phi} \prod_{j=i}^{i+6} \left[ TN_j < TN10_j \right]

    where :math:`[P]` is 1 if :math:`P` is true, and 0 if false.

    References
    ----------
    From the Expert Team on Climate Change Detection, Monitoring and Indices (ETCCDMI; :cite:p:`zhang_indices_2011`).

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import cold_spell_duration_index
    >>> tasmin = xr.open_dataset(path_to_tasmin_file).tasmin.isel(lat=0, lon=0)
    >>> tn10 = percentile_doy(tasmin, per=10).sel(percentiles=10)
    >>> cold_spell_duration_index(tasmin, tn10)

    Note that this example does not use a proper 1961-1990 reference period.
    """
    tasmin_per = convert_units_to(tasmin_per, tasmin)

    # Create time series out of doy values.
    thresh = resample_doy(tasmin_per, tasmin)

    below = compare(tasmin, op, thresh, constrain=("<", "<="))
    out = rl.resample_and_rl(
        below,
        resample_before_rl,
        rl.windowed_run_count,
        window=window,
        freq=freq,
    )

    return to_agg_units(out, tasmin, "count")


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    tas_per="[temperature]",
    pr_per="[precipitation]",
)
def cold_and_dry_days(
    tas: xarray.DataArray,
    pr: xarray.DataArray,
    tas_per: xarray.DataArray,
    pr_per: xarray.DataArray,
    freq: str = "YS",
) -> xarray.DataArray:
    r"""Cold and dry days.

    Returns the total number of days when "Cold" and "Dry" conditions coincide.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature values
    pr : xarray.DataArray
        Daily precipitation.
    tas_per : xarray.DataArray
        First quartile of daily mean temperature computed by month.
    pr_per : xarray.DataArray
        First quartile of daily total precipitation computed by month.
    freq : str
        Resampling frequency.

    Warnings
    --------
    Before computing the percentiles, all the precipitation below 1mm must be filtered out!
    Otherwise, the percentiles will include non-wet days.

    Returns
    -------
    xarray.DataArray
        The total number of days when cold and dry conditions coincide.

    Notes
    -----
    Bootstrapping is not available for quartiles because it would make no significant difference to bootstrap
    percentiles so far from the extremes.

    Formula to be written (:cite:t:`beniston_trends_2009`)

    References
    ----------
    :cite:cts:`beniston_trends_2009`

    """
    tas_per = convert_units_to(tas_per, tas)
    thresh = resample_doy(tas_per, tas)
    tg25 = tas < thresh

    pr_per = convert_units_to(pr_per, pr, context="hydro")
    thresh = resample_doy(pr_per, pr)
    pr25 = pr < thresh

    cold_and_dry = cast(xarray.DataArray, np.logical_and(tg25, pr25))
    resampled = cold_and_dry.resample(time=freq).sum(dim="time")
    out = to_agg_units(resampled, tas, "count")
    return out


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    tas_per="[temperature]",
    pr_per="[precipitation]",
)
def warm_and_dry_days(
    tas: xarray.DataArray,
    pr: xarray.DataArray,
    tas_per: xarray.DataArray,
    pr_per: xarray.DataArray,
    freq: str = "YS",
) -> xarray.DataArray:
    r"""Warm and dry days.

    Returns the total number of days when "warm" and "Dry" conditions coincide.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature values
    pr : xarray.DataArray
        Daily precipitation.
    tas_per : xarray.DataArray
        Third quartile of daily mean temperature computed by month.
    pr_per : xarray.DataArray
        First quartile of daily total precipitation computed by month.
    freq : str
        Resampling frequency.

    Warnings
    --------
    Before computing the percentiles, all the precipitation below 1mm must be filtered out!
    Otherwise, the percentiles will include non-wet days.

    Returns
    -------
    xarray.DataArray,
        The total number of days when warm and dry conditions coincide.

    Notes
    -----
    Bootstrapping is not available for quartiles because it would make no significant difference to bootstrap
    percentiles so far from the extremes.

    Formula to be written (:cite:t:`beniston_trends_2009`)

    References
    ----------
    :cite:cts:`beniston_trends_2009`

    """
    tas_per = convert_units_to(tas_per, tas)
    thresh = resample_doy(tas_per, tas)
    tg75 = tas > thresh

    pr_per = convert_units_to(pr_per, pr, context="hydro")
    thresh = resample_doy(pr_per, pr)
    pr25 = pr < thresh

    warm_and_dry = cast(xarray.DataArray, np.logical_and(tg75, pr25))
    resampled = warm_and_dry.resample(time=freq).sum(dim="time")
    out = to_agg_units(resampled, tas, "count")
    return out


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    tas_per="[temperature]",
    pr_per="[precipitation]",
)
def warm_and_wet_days(
    tas: xarray.DataArray,
    pr: xarray.DataArray,
    tas_per: xarray.DataArray,
    pr_per: xarray.DataArray,
    freq: str = "YS",
) -> xarray.DataArray:
    r"""Warm and wet days.

    Returns the total number of days when "warm" and "wet" conditions coincide.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature values
    pr : xarray.DataArray
        Daily precipitation.
    tas_per : xarray.DataArray
        Third quartile of daily mean temperature computed by month.
    pr_per : xarray.DataArray
        Third quartile of daily total precipitation computed by month.
    freq : str
        Resampling frequency.

    Warnings
    --------
    Before computing the percentiles, all the precipitation below 1mm must be filtered out!
    Otherwise, the percentiles will include non-wet days.

    Returns
    -------
    xarray.DataArray
        The total number of days when warm and wet conditions coincide.

    Notes
    -----
    Bootstrapping is not available for quartiles because it would make no significant difference
    to bootstrap percentiles so far from the extremes.

    Formula to be written (:cite:t:`beniston_trends_2009`)

    References
    ----------
    :cite:cts:`beniston_trends_2009`
    """
    tas_per = convert_units_to(tas_per, tas)
    thresh = resample_doy(tas_per, tas)
    tg75 = tas > thresh

    pr_per = convert_units_to(pr_per, pr, context="hydro")
    thresh = resample_doy(pr_per, pr)
    pr75 = pr > thresh

    warm_and_wet = cast(xarray.DataArray, np.logical_and(tg75, pr75))
    resampled = warm_and_wet.resample(time=freq).sum(dim="time")
    out = to_agg_units(resampled, tas, "count")
    return out


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    tas_per="[temperature]",
    pr_per="[precipitation]",
)
def cold_and_wet_days(
    tas: xarray.DataArray,
    pr: xarray.DataArray,
    tas_per: xarray.DataArray,
    pr_per: xarray.DataArray,
    freq: str = "YS",
) -> xarray.DataArray:
    r"""Cold and wet days.

    Returns the total number of days when "cold" and "wet" conditions coincide.

    Warnings
    --------
    Before computing the percentiles, all the precipitation below 1mm must be filtered out!
    Otherwise, the percentiles will include non-wet days.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature values
    pr : xarray.DataArray
        Daily precipitation.
    tas_per : xarray.DataArray
        First quartile of daily mean temperature computed by month.
    pr_per : xarray.DataArray
        Third quartile of daily total precipitation computed by month.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The total number of days when cold and wet conditions coincide.

    Notes
    -----
    Bootstrapping is not available for quartiles because it would make no significant
    difference to bootstrap percentiles so far from the extremes.

    Formula to be written (:cite:t:`beniston_trends_2009`)

    References
    ----------
    :cite:cts:`beniston_trends_2009`
    """
    tas_per = convert_units_to(tas_per, tas)
    thresh = resample_doy(tas_per, tas)
    tg25 = tas < thresh

    pr_per = convert_units_to(pr_per, pr, context="hydro")
    thresh = resample_doy(pr_per, pr)
    pr75 = pr > thresh

    cold_and_wet = cast(xarray.DataArray, np.logical_and(tg25, pr75))
    resampled = cold_and_wet.resample(time=freq).sum(dim="time")
    out = to_agg_units(resampled, tas, "count")
    return out


@declare_units(
    tasmin="[temperature]",
    tasmax="[temperature]",
    thresh_tasmin="[temperature]",
    thresh_tasmax="[temperature]",
)
def multiday_temperature_swing(
    tasmin: xarray.DataArray,
    tasmax: xarray.DataArray,
    thresh_tasmin: Quantified = "0 degC",
    thresh_tasmax: Quantified = "0 degC",
    window: int = 1,
    op: str = "mean",
    op_tasmin: str = "<=",
    op_tasmax: str = ">",
    freq: str = "YS",
    resample_before_rl: bool = True,
) -> xarray.DataArray:
    r"""Statistics of consecutive diurnal temperature swing events.

    A diurnal swing of max and min temperature event is when Tmax > thresh_tasmax and Tmin <= thresh_tasmin. This indice
    finds all days that constitute these events and computes statistics over the length and frequency of these events.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh_tasmin : Quantified
        The temperature threshold needed to trigger a freeze event.
    thresh_tasmax : Quantified
        The temperature threshold needed to trigger a thaw event.
    window : int
        The minimal length of spells to be included in the statistics.
    op : {'mean', 'sum', 'max', 'min', 'std', 'count'}
        The statistical operation to use when reducing the list of spell lengths.
    op_tasmin : {"<", "<=", "lt", "le"}
        Comparison operation for tasmin. Default: "<=".
    op_tasmax : {">", ">=", "gt", "ge"}
        Comparison operation for tasmax. Default: ">".
    freq : str
        Resampling frequency.
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.

    Returns
    -------
    xarray.DataArray, [time]
        {freq} {op} length of diurnal temperature cycles exceeding thresholds.

    Notes
    -----
    Let :math:`TX_{i}` be the maximum temperature at day :math:`i` and :math:`TN_{i}` be the daily minimum temperature
    at day :math:`i`. Then freeze thaw spells during a given period are consecutive days where:

    .. math::

        TX_{i} > 0℃ \land TN_{i} <  0℃

    This indice returns a given statistic of the found lengths, optionally dropping those shorter than the `window`
    argument. For example, `window=1` and `op='sum'` returns the same value as :py:func:`daily_freezethaw_cycles`.
    """
    thaw_threshold = convert_units_to(thresh_tasmax, tasmax)
    freeze_threshold = convert_units_to(thresh_tasmin, tasmin)

    freeze = compare(tasmin, op_tasmin, freeze_threshold, constrain=("<", "<="))
    thaw = compare(tasmax, op_tasmax, thaw_threshold, constrain=(">", ">="))
    ft = freeze * thaw

    if op == "count":
        out = rl.resample_and_rl(
            ft,
            resample_before_rl,
            rl.windowed_run_events,
            window=window,
            freq=freq,
        )
    else:
        out = rl.resample_and_rl(
            ft,
            resample_before_rl,
            rl.rle_statistics,
            reducer=op,
            window=window,
            freq=freq,
        )

    return to_agg_units(out, tasmin, "count")


@declare_units(tasmax="[temperature]", tasmin="[temperature]")
def daily_temperature_range(
    tasmin: xarray.DataArray,
    tasmax: xarray.DataArray,
    freq: str = "YS",
    op: str | Callable = "mean",
) -> xarray.DataArray:
    r"""Statistics of daily temperature range.

    The mean difference between the daily maximum temperature and the daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    freq : str
        Resampling frequency.
    op : {'min', 'max', 'mean', 'std'} or func
        Reduce operation. Can either be a DataArray method or a function that can be applied to a DataArray.

    Returns
    -------
    xarray.DataArray, [same units as tasmin]
        The average variation in daily temperature range for the given time period.

    Notes
    -----
    For a default calculation using `op='mean'`:

    Let :math:`TX_{ij}` and :math:`TN_{ij}` be the daily maximum and minimum temperature at day :math:`i` of period
    :math:`j`. Then the mean diurnal temperature range in period :math:`j` is:

    .. math::

        DTR_j = \frac{ \sum_{i=1}^I (TX_{ij} - TN_{ij}) }{I}
    """
    tasmax = convert_units_to(tasmax, tasmin)
    dtr = tasmax - tasmin
    u = str2pint(tasmax.units)
    dtr.attrs.update(pint2cfattrs(u, is_difference=True))
    out = select_resample_op(dtr, op=op, freq=freq)
    return out


@declare_units(tasmax="[temperature]", tasmin="[temperature]")
def daily_temperature_range_variability(
    tasmin: xarray.DataArray, tasmax: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""Mean absolute day-to-day variation in daily temperature range.

    Mean absolute day-to-day variation in daily temperature range.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmin]
        The average day-to-day variation in daily temperature range for the given time period.

    Notes
    -----
    Let :math:`TX_{ij}` and :math:`TN_{ij}` be the daily maximum and minimum temperature at day :math:`i` of period
    :math:`j`. Then calculated is the absolute day-to-day differences in period :math:`j` is:

    .. math::

       vDTR_j = \frac{ \sum_{i=2}^{I} |(TX_{ij}-TN_{ij})-(TX_{i-1,j}-TN_{i-1,j})| }{I}
    """
    tasmax = convert_units_to(tasmax, tasmin)
    vdtr = abs((tasmax - tasmin).diff(dim="time"))
    u = str2pint(tasmax.units)
    vdtr.attrs.update(pint2cfattrs(u, is_difference=True))
    out = select_resample_op(vdtr, op="mean", freq=freq)
    return out


@declare_units(tasmax="[temperature]", tasmin="[temperature]")
def extreme_temperature_range(
    tasmin: xarray.DataArray, tasmax: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""Extreme intra-period temperature range.

    The maximum of max temperature (TXx) minus the minimum of min temperature (TNn) for the given time period.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmin]
        Extreme intra-period temperature range for the given time period.

    Notes
    -----
    Let :math:`TX_{ij}` and :math:`TN_{ij}` be the daily maximum and minimum temperature at day :math:`i` of period
    :math:`j`. Then the extreme temperature range in period :math:`j` is:

    .. math::

        ETR_j = max(TX_{ij}) - min(TN_{ij})
    """
    tasmax = convert_units_to(tasmax, tasmin)
    tx_max = tasmax.resample(time=freq).max(dim="time")
    tn_min = tasmin.resample(time=freq).min(dim="time")

    out = tx_max - tn_min
    u = str2pint(tasmax.units)
    out.attrs.update(pint2cfattrs(u, is_difference=True))
    return out


@declare_units(
    tasmin="[temperature]",
    tasmax="[temperature]",
    thresh_tasmin="[temperature]",
    thresh_tasmax="[temperature]",
)
def heat_wave_frequency(
    tasmin: xarray.DataArray,
    tasmax: xarray.DataArray,
    thresh_tasmin: Quantified = "22.0 degC",
    thresh_tasmax: Quantified = "30 degC",
    window: int = 3,
    freq: str = "YS",
    op: str = ">",
    resample_before_rl: bool = True,
) -> xarray.DataArray:
    r"""Heat wave frequency.

    Number of heat waves over a given period. A heat wave is defined as an event where the minimum and maximum daily
    temperature both exceed specific thresholds over a minimum number of days.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh_tasmin : Quantified
        The minimum temperature threshold needed to trigger a heatwave event.
    thresh_tasmax : Quantified
        The maximum temperature threshold needed to trigger a heatwave event.
    window : int
        Minimum number of days with temperatures above thresholds to qualify as a heatwave.
    freq : str
        Resampling frequency.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Number of heatwave at the requested frequency.

    Notes
    -----
    The thresholds of 22° and 25°C for night temperatures and 30° and 35°C for day temperatures were selected by
    Health Canada professionals, following a temperature–mortality analysis. These absolute temperature thresholds
    characterize the occurrence of hot weather events that can result in adverse health outcomes for Canadian
    communities :cite:p:`casati_regional_2013`.

    In :cite:t:`robinson_definition_2001`, the parameters would be `thresh_tasmin=27.22, thresh_tasmax=39.44, window=2` (81F, 103F).

    References
    ----------
    :cite:cts:`casati_regional_2013,robinson_definition_2001`
    """
    thresh_tasmax = convert_units_to(thresh_tasmax, tasmax)
    thresh_tasmin = convert_units_to(thresh_tasmin, tasmin)

    constrain = (">", ">=")
    cond = (compare(tasmin, op, thresh_tasmin, constrain)) & (
        compare(tasmax, op, thresh_tasmax, constrain)
    )

    out = rl.resample_and_rl(
        cond,
        resample_before_rl,
        rl.windowed_run_events,
        window=window,
        freq=freq,
    )
    out = out.assign_attrs(units="")
    return out


@declare_units(
    tasmin="[temperature]",
    tasmax="[temperature]",
    thresh_tasmin="[temperature]",
    thresh_tasmax="[temperature]",
)
def heat_wave_max_length(
    tasmin: xarray.DataArray,
    tasmax: xarray.DataArray,
    thresh_tasmin: Quantified = "22.0 degC",
    thresh_tasmax: Quantified = "30 degC",
    window: int = 3,
    freq: str = "YS",
    op: str = ">",
    resample_before_rl: bool = True,
) -> xarray.DataArray:
    r"""Heat wave max length.

    Maximum length of heat waves over a given period. A heat wave is defined as an event where the minimum and maximum
    daily temperature both exceeds specific thresholds over a minimum number of days.

    By definition heat_wave_max_length must be >= window.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh_tasmin : Quantified
        The minimum temperature threshold needed to trigger a heatwave event.
    thresh_tasmax : Quantified
        The maximum temperature threshold needed to trigger a heatwave event.
    window : int
        Minimum number of days with temperatures above thresholds to qualify as a heatwave.
    freq : str
        Resampling frequency.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.

    Returns
    -------
    xarray.DataArray, [time]
        Maximum length of heatwave at the requested frequency.

    Notes
    -----
    The thresholds of 22° and 25°C for night temperatures and 30° and 35°C for day temperatures were selected by
    Health Canada professionals, following a temperature–mortality analysis. These absolute temperature thresholds
    characterize the occurrence of hot weather events that can result in adverse health outcomes for Canadian
    communities :cite:p:`casati_regional_2013`.

    In :cite:t:`robinson_definition_2001`, the parameters would be:
    `thresh_tasmin=27.22, thresh_tasmax=39.44, window=2` (81F, 103F).

    References
    ----------
    :cite:cts:`casati_regional_2013,robinson_definition_2001`
    """
    thresh_tasmax = convert_units_to(thresh_tasmax, tasmax)
    thresh_tasmin = convert_units_to(thresh_tasmin, tasmin)

    constrain = (">", ">=")
    cond = (compare(tasmin, op, thresh_tasmin, constrain)) & (
        compare(tasmax, op, thresh_tasmax, constrain)
    )
    out = rl.resample_and_rl(
        cond,
        resample_before_rl,
        rl.rle_statistics,
        reducer="max",
        window=window,
        freq=freq,
    )
    return to_agg_units(out, tasmax, "count")


@declare_units(
    tasmin="[temperature]",
    tasmax="[temperature]",
    thresh_tasmin="[temperature]",
    thresh_tasmax="[temperature]",
)
def heat_wave_total_length(
    tasmin: xarray.DataArray,
    tasmax: xarray.DataArray,
    thresh_tasmin: Quantified = "22.0 degC",
    thresh_tasmax: Quantified = "30 degC",
    window: int = 3,
    freq: str = "YS",
    op: str = ">",
    resample_before_rl: bool = True,
) -> xarray.DataArray:
    r"""Heat wave total length.

    Total length of heat waves over a given period. A heat wave is defined as an event where the minimum and maximum
    daily temperature both exceeds specific thresholds over a minimum number of days.
    This the sum of all days in such events.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh_tasmin : str
        The minimum temperature threshold needed to trigger a heatwave event.
    thresh_tasmax : str
        The maximum temperature threshold needed to trigger a heatwave event.
    window : int
        Minimum number of days with temperatures above thresholds to qualify as a heatwave.
    freq : str
        Resampling frequency.
    op: {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.

    Returns
    -------
    xarray.DataArray, [time]
        Total length of heatwave at the requested frequency.

    Notes
    -----
    See notes and references of `heat_wave_max_length`
    """
    thresh_tasmax = convert_units_to(thresh_tasmax, tasmax)
    thresh_tasmin = convert_units_to(thresh_tasmin, tasmin)

    constrain = (">", ">=")
    cond = compare(tasmin, op, thresh_tasmin, constrain) & compare(
        tasmax, op, thresh_tasmax, constrain
    )
    out = rl.resample_and_rl(
        cond,
        resample_before_rl,
        rl.windowed_run_count,
        window=window,
        freq=freq,
    )

    return to_agg_units(out, tasmin, "count")


@declare_units(
    pr="[precipitation]",
    prsn="[precipitation]",
    tas="[temperature]",
    thresh="[temperature]",
)
def liquid_precip_ratio(
    pr: xarray.DataArray,
    prsn: xarray.DataArray | None = None,
    tas: xarray.DataArray | None = None,
    thresh: Quantified = "0 degC",
    freq: str = "QS-DEC",
) -> xarray.DataArray:
    r"""Ratio of rainfall to total precipitation.

    The ratio of total liquid precipitation over the total precipitation. If solid precipitation is not provided,
    it is approximated with pr, tas and thresh, using the `snowfall_approximation` function with method 'binary'.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    prsn : xarray.DataArray, optional
        Mean daily solid precipitation flux.
    tas : xarray.DataArray, optional
        Mean daily temperature.
    thresh : Quantified
        Threshold temperature under which precipitation is assumed to be solid.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Ratio of rainfall to total precipitation.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation of day :math:`i`, then for a period :math:`j` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

        PR_{ij} = \sum_{i=a}^{b} PR_i

        PRwet_{ij}

    See Also
    --------
    winter_rain_ratio
    """
    if prsn is None and tas is not None:
        prsn = snowfall_approximation(pr, tas=tas, thresh=thresh, method="binary")
    elif prsn is None:
        raise KeyError("prsn or tas must be supplied.")

    tot = pr.resample(time=freq).sum(dim="time")
    rain = tot - prsn.resample(time=freq).sum(dim="time")
    ratio = rain / tot
    ratio = ratio.assign_attrs(units="")
    return ratio


@declare_units(pr="[precipitation]", tas="[temperature]", thresh="[temperature]")
def precip_accumulation(
    pr: xarray.DataArray,
    tas: xarray.DataArray | None = None,
    phase: str | None = None,
    thresh: Quantified = "0 degC",
    freq: str = "YS",
) -> xarray.DataArray:
    r"""Accumulated total (liquid and/or solid) precipitation.

    Resample the original daily mean precipitation flux and accumulate over each period.
    If a daily temperature is provided, the `phase` keyword can be used to sum precipitation of a given phase only.
    When the temperature is under the given threshold, precipitation is assumed to be snow, and liquid rain otherwise.
    This indice is agnostic to the type of daily temperature (tas, tasmax or tasmin) given.

    Parameters
    ----------
    pr : xarray.DataArray
      Mean daily precipitation flux.
    tas : xarray.DataArray, optional
      Mean, maximum or minimum daily temperature.
    phase : {None, 'liquid', 'solid'}
      Which phase to consider, "liquid" or "solid", if None (default), both are considered.
    thresh : Quantified
      Threshold of `tas` over which the precipication is assumed to be liquid rain.
    freq : str
      Resampling frequency.

    Returns
    -------
    xarray.DataArray, [length]
      The total daily precipitation at the given time frequency for the given phase.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation of day :math:`i`, then for a period :math:`j` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       PR_{ij} = \sum_{i=a}^{b} PR_i

    If tas and phase are given, the corresponding phase precipitation is estimated before computing the accumulation,
    using one of `snowfall_approximation` or `rain_approximation` with the `binary` method.

    Examples
    --------
    The following would compute, for each grid cell of a dataset, the total
    precipitation at the seasonal frequency, ie DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import precip_accumulation
    >>> pr_day = xr.open_dataset(path_to_pr_file).pr
    >>> prcp_tot_seasonal = precip_accumulation(pr_day, freq="QS-DEC")
    """
    if phase == "liquid":
        pr = rain_approximation(pr, tas=tas, thresh=thresh, method="binary")
    elif phase == "solid":
        pr = snowfall_approximation(pr, tas=tas, thresh=thresh, method="binary")
    pram = rate2amount(pr)
    pram = pram.resample(time=freq).sum(dim="time").assign_attrs(units=pram.units)
    return pram


@declare_units(pr="[precipitation]", tas="[temperature]", thresh="[temperature]")
def precip_average(
    pr: xarray.DataArray,
    tas: xarray.DataArray | None = None,
    phase: str | None = None,
    thresh: Quantified = "0 degC",
    freq: str = "YS",
) -> xarray.DataArray:
    r"""Averaged (liquid and/or solid) precipitation.

    Resample the original daily mean precipitation flux and average over each period.
    If a daily temperature is provided, the `phase` keyword can be used to average precipitation of a given phase only.
    When the temperature is under the given threshold, precipitation is assumed to be snow, and liquid rain otherwise.
    This indice is agnostic to the type of daily temperature (tas, tasmax or tasmin) given.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    tas : xarray.DataArray, optional
        Mean, maximum or minimum daily temperature.
    phase : {None, 'liquid', 'solid'}
        Which phase to consider, "liquid" or "solid", if None (default), both are considered.
    thresh : Quantified
        Threshold of `tas` over which the precipication is assumed to be liquid rain.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [length]
        The averaged daily precipitation at the given time frequency for the given phase.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation of day :math:`i`, then for a period :math:`j` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       PR_{ij} =\frac{ \sum_{i=a}^{b} PR_i }{b - a + 1}

    If tas and phase are given, the corresponding phase precipitation is estimated before computing the accumulation,
    using one of `snowfall_approximation` or `rain_approximation` with the `binary` method.

    Examples
    --------
    The following would compute, for each grid cell of a dataset, the total
    precipitation at the seasonal frequency, ie DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import precip_average
    >>> pr_day = xr.open_dataset(path_to_pr_file).pr
    >>> prcp_tot_seasonal = precip_average(pr_day, freq="QS-DEC")
    """
    if phase == "liquid":
        pr = rain_approximation(pr, tas=tas, thresh=thresh, method="binary")
    elif phase == "solid":
        pr = snowfall_approximation(pr, tas=tas, thresh=thresh, method="binary")
    pram = rate2amount(pr)
    pram = pram.resample(time=freq).mean(dim="time").assign_attrs(units=pram.units)
    return pram


# FIXME: Resample after run length?
@declare_units(pr="[precipitation]", tas="[temperature]", thresh="[precipitation]")
def rain_on_frozen_ground_days(
    pr: xarray.DataArray,
    tas: xarray.DataArray,
    thresh: Quantified = "1 mm/d",
    freq: str = "YS",
) -> xarray.DataArray:
    """Number of rain on frozen ground events.

    Number of days with rain above a threshold after a series of seven days below freezing temperature.
    Precipitation is assumed to be rain when the temperature is above 0℃.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    tas : xarray.DataArray
        Mean daily temperature.
    thresh : Quantified
        Precipitation threshold to consider a day as a rain event.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [time]
        The number of rain on frozen ground events per period.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation and :math:`TG_i` be the mean daily temperature of day :math:`i`.
    Then for a period :math:`j`, rain on frozen grounds days are counted where:

    .. math::

        PR_{i} > Threshold [mm]

    and where

    .. math::

        TG_{i} ≤ 0℃

    is true for continuous periods where :math:`i ≥ 7`
    """
    t = convert_units_to(thresh, pr, context="hydro")
    frz = convert_units_to("0 C", tas)

    def func(x, axis):
        """Check that temperature conditions are below 0 for seven days and above after."""
        frozen = x == np.array([0, 0, 0, 0, 0, 0, 0, 1], bool)
        return frozen.all(axis=axis)

    tcond = (tas > frz).rolling(time=8).reduce(func)
    pcond = pr > t

    out = (tcond * pcond * 1).resample(time=freq).sum(dim="time")
    return to_agg_units(out, tas, "count")


@declare_units(
    pr="[precipitation]",
    tas="[temperature]",
    pr_thresh="[precipitation]",
    tas_thresh="[temperature]",
)
def high_precip_low_temp(
    pr: xarray.DataArray,
    tas: xarray.DataArray,
    pr_thresh: Quantified = "0.4 mm/d",
    tas_thresh: Quantified = "-0.2 degC",
    freq: str = "YS",
) -> xarray.DataArray:
    """Number of days with precipitation above threshold and temperature below threshold.

    Number of days when precipitation is greater or equal to some threshold, and temperatures are colder than some
    threshold. This can be used for example to identify days with the potential for freezing rain or icing conditions.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    tas : xarray.DataArray
        Daily mean, minimum or maximum temperature.
    pr_thresh : Quantified
        Precipitation threshold to exceed.
    tas_thresh : Quantified
        Temperature threshold not to exceed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with high precipitation and low temperatures.

    Example
    -------
    To compute the number of days with intense rainfall while minimum temperatures dip below -0.2C:
    >>> pr = xr.open_dataset(path_to_pr_file).pr
    >>> tasmin = xr.open_dataset(path_to_tasmin_file).tasmin
    >>> high_precip_low_temp(
    ...     pr, tas=tasmin, pr_thresh="10 mm/d", tas_thresh="-0.2 degC"
    ... )
    """
    pr_thresh = convert_units_to(pr_thresh, pr, context="hydro")
    tas_thresh = convert_units_to(tas_thresh, tas)

    cond = (pr >= pr_thresh) * (tas < tas_thresh) * 1
    out = cond.resample(time=freq).sum(dim="time")
    return to_agg_units(out, pr, "count")


@declare_units(pr="[precipitation]", pr_per="[precipitation]", thresh="[precipitation]")
@percentile_bootstrap
def days_over_precip_thresh(
    pr: xarray.DataArray,
    pr_per: xarray.DataArray,
    thresh: Quantified = "1 mm/day",
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = ">",
) -> xarray.DataArray:
    r"""Number of wet days with daily precipitation over a given percentile.

    Number of days over period where the precipitation is above a threshold defining wet days and above a given
    percentile for that day.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    pr_per : xarray.DataArray
        Percentile of wet day precipitation flux. Either computed daily (one value per day
        of year) or computed over a period (one value per spatial point).
    thresh : Quantified
        Precipitation value over which a day is considered wet.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [time]
      Count of days with daily precipitation above the given percentile [days].

    Examples
    --------
    >>> from xclim.indices import days_over_precip_thresh
    >>> pr = xr.open_dataset(path_to_pr_file).pr
    >>> p75 = pr.quantile(0.75, dim="time", keep_attrs=True)
    >>> r75p = days_over_precip_thresh(pr, p75)
    """
    pr_per = convert_units_to(pr_per, pr, context="hydro")
    thresh = convert_units_to(thresh, pr, context="hydro")

    tp = pr_per.where(pr_per > thresh, thresh)
    if "dayofyear" in pr_per.coords:
        # Create time series out of doy values.
        tp = resample_doy(tp, pr)

    # Compute the days when precip is both over the wet day threshold and the percentile threshold.
    out = threshold_count(pr, op, tp, freq, constrain=(">", ">="))
    return to_agg_units(out, pr, "count")


@declare_units(pr="[precipitation]", pr_per="[precipitation]", thresh="[precipitation]")
@percentile_bootstrap
def fraction_over_precip_thresh(
    pr: xarray.DataArray,
    pr_per: xarray.DataArray,
    thresh: Quantified = "1 mm/day",
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = ">",
) -> xarray.DataArray:
    r"""Fraction of precipitation due to wet days with daily precipitation over a given percentile.

    Percentage of the total precipitation over period occurring in days when the precipitation is above a threshold
    defining wet days and above a given percentile for that day.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    pr_per : xarray.DataArray
        Percentile of wet day precipitation flux. Either computed daily (one value per day
        of year) or computed over a period (one value per spatial point).
    thresh : Quantified
        Precipitation value over which a day is considered wet.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Fraction of precipitation over threshold during wet days.

    """
    pr_per = convert_units_to(pr_per, pr, context="hydro")
    thresh = convert_units_to(thresh, pr, context="hydro")

    tp = pr_per.where(pr_per > thresh, thresh)
    if "dayofyear" in pr_per.coords:
        # Create time series out of doy values.
        tp = resample_doy(tp, pr)

    constrain = (">", ">=")
    # Total precip during wet days over period
    total = (
        pr.where(compare(pr, op, thresh, constrain), 0)
        .resample(time=freq)
        .sum(dim="time")
    )

    # Compute the days when precip is both over the wet day threshold and the percentile threshold.
    over = (
        pr.where(compare(pr, op, tp, constrain), 0).resample(time=freq).sum(dim="time")
    )

    out = over / total
    out.attrs["units"] = ""
    return out


@declare_units(tas="[temperature]", tas_per="[temperature]")
@percentile_bootstrap
def tg90p(
    tas: xarray.DataArray,
    tas_per: xarray.DataArray,
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = ">",
) -> xarray.DataArray:
    r"""Number of days with daily mean temperature over the 90th percentile.

    Number of days with daily mean temperature over the 90th percentile.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature.
    tas_per : xarray.DataArray
        90th percentile of daily mean temperature.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with daily mean temperature below the 10th percentile [days].

    Notes
    -----
    The 90th percentile should be computed for a 5-day window centered on each calendar day for a reference period.

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import tg90p
    >>> tas = xr.open_dataset(path_to_tas_file).tas
    >>> tas_per = percentile_doy(tas, per=90).sel(percentiles=90)
    >>> hot_days = tg90p(tas, tas_per)
    """
    tas_per = convert_units_to(tas_per, tas)

    # Create time series out of doy values.
    thresh = resample_doy(tas_per, tas)

    # Identify the days over the 90th percentile
    out = threshold_count(tas, op, thresh, freq, constrain=(">", ">="))
    return to_agg_units(out, tas, "count")


@declare_units(tas="[temperature]", tas_per="[temperature]")
@percentile_bootstrap
def tg10p(
    tas: xarray.DataArray,
    tas_per: xarray.DataArray,
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = "<",
) -> xarray.DataArray:
    r"""Number of days with daily mean temperature below the 10th percentile.

    Number of days with daily mean temperature below the 10th percentile.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature.
    tas_per : xarray.DataArray
        10th percentile of daily mean temperature.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {"<", "<=", "lt", "le"}
        Comparison operation. Default: "<".

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with daily mean temperature below the 10th percentile [days].

    Notes
    -----
    The 10th percentile should be computed for a 5-day window centered on each calendar day for a reference period.

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import tg10p
    >>> tas = xr.open_dataset(path_to_tas_file).tas
    >>> tas_per = percentile_doy(tas, per=10).sel(percentiles=10)
    >>> cold_days = tg10p(tas, tas_per)
    """
    tas_per = convert_units_to(tas_per, tas)

    # Create time series out of doy values.
    thresh = resample_doy(tas_per, tas)

    # Identify the days below the 10th percentile
    out = threshold_count(tas, op, thresh, freq, constrain=("<", "<="))
    return to_agg_units(out, tas, "count")


@declare_units(tasmin="[temperature]", tasmin_per="[temperature]")
@percentile_bootstrap
def tn90p(
    tasmin: xarray.DataArray,
    tasmin_per: xarray.DataArray,
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = ">",
) -> xarray.DataArray:
    r"""Number of days with daily minimum temperature over the 90th percentile.

    Number of days with daily minimum temperature over the 90th percentile.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmin_per : xarray.DataArray
        90th percentile of daily minimum temperature.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with daily minimum temperature below the 10th percentile [days].

    Notes
    -----
    The 90th percentile should be computed for a 5-day window centered on each calendar day for a reference period.

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import tn90p
    >>> tas = xr.open_dataset(path_to_tas_file).tas
    >>> tas_per = percentile_doy(tas, per=90).sel(percentiles=90)
    >>> hot_days = tn90p(tas, tas_per)
    """
    tasmin_per = convert_units_to(tasmin_per, tasmin)

    # Create time series out of doy values.
    thresh = resample_doy(tasmin_per, tasmin)

    # Identify the days with min temp above 90th percentile.
    out = threshold_count(tasmin, op, thresh, freq, constrain=(">", ">="))
    return to_agg_units(out, tasmin, "count")


@declare_units(tasmin="[temperature]", tasmin_per="[temperature]")
@percentile_bootstrap
def tn10p(
    tasmin: xarray.DataArray,
    tasmin_per: xarray.DataArray,
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = "<",
) -> xarray.DataArray:
    r"""Number of days with daily minimum temperature below the 10th percentile.

    Number of days with daily minimum temperature below the 10th percentile.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Mean daily temperature.
    tasmin_per : xarray.DataArray
        10th percentile of daily minimum temperature.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {"<", "<=", "lt", "le"}
        Comparison operation. Default: "<".

    Returns
    -------
    xarray.DataArray, [time]
      Count of days with daily minimum temperature below the 10th percentile [days].

    Notes
    -----
    The 10th percentile should be computed for a 5-day window centered on each calendar day for a reference period.

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import tn10p
    >>> tas = xr.open_dataset(path_to_tas_file).tas
    >>> tas_per = percentile_doy(tas, per=10).sel(percentiles=10)
    >>> cold_days = tn10p(tas, tas_per)
    """
    tasmin_per = convert_units_to(tasmin_per, tasmin)

    # Create time series out of doy values.
    thresh = resample_doy(tasmin_per, tasmin)

    # Identify the days below the 10th percentile
    out = threshold_count(tasmin, op, thresh, freq, constrain=("<", "<="))
    return to_agg_units(out, tasmin, "count")


@declare_units(tasmax="[temperature]", tasmax_per="[temperature]")
@percentile_bootstrap
def tx90p(
    tasmax: xarray.DataArray,
    tasmax_per: xarray.DataArray,
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = ">",
) -> xarray.DataArray:
    r"""Number of days with daily maximum temperature over the 90th percentile.

    Number of days with daily maximum temperature over the 90th percentile.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    tasmax_per : xarray.DataArray
        90th percentile of daily maximum temperature.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with daily maximum temperature below the 10th percentile [days].

    Notes
    -----
    The 90th percentile should be computed for a 5-day window centered on each calendar day for a reference period.

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import tx90p
    >>> tas = xr.open_dataset(path_to_tas_file).tas
    >>> tasmax_per = percentile_doy(tas, per=90).sel(percentiles=90)
    >>> hot_days = tx90p(tas, tasmax_per)
    """
    tasmax_per = convert_units_to(tasmax_per, tasmax)

    # Create time series out of doy values.
    thresh = resample_doy(tasmax_per, tasmax)

    # Identify the days with max temp above 90th percentile.
    out = threshold_count(tasmax, op, thresh, freq, constrain=(">", ">="))
    return to_agg_units(out, tasmax, "count")


@declare_units(tasmax="[temperature]", tasmax_per="[temperature]")
@percentile_bootstrap
def tx10p(
    tasmax: xarray.DataArray,
    tasmax_per: xarray.DataArray,
    freq: str = "YS",
    bootstrap: bool = False,  # noqa
    op: str = "<",
) -> xarray.DataArray:
    r"""Number of days with daily maximum temperature below the 10th percentile.

    Number of days with daily maximum temperature below the 10th percentile.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    tasmax_per : xarray.DataArray
        10th percentile of daily maximum temperature.
    freq : str
        Resampling frequency.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {"<", "<=", "lt", "le"}
        Comparison operation. Default: "<".

    Returns
    -------
    xarray.DataArray, [time]
        Count of days with daily maximum temperature below the 10th percentile [days].

    Notes
    -----
    The 10th percentile should be computed for a 5-day window centered on each calendar day for a reference period.

    Examples
    --------
    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import tx10p
    >>> tas = xr.open_dataset(path_to_tas_file).tas
    >>> tasmax_per = percentile_doy(tas, per=10).sel(percentiles=10)
    >>> cold_days = tx10p(tas, tasmax_per)
    """
    tasmax_per = convert_units_to(tasmax_per, tasmax)

    # Create time series out of doy values.
    thresh = resample_doy(tasmax_per, tasmax)

    # Identify the days below the 10th percentile
    out = threshold_count(tasmax, op, thresh, freq, constrain=("<", "<="))
    return to_agg_units(out, tasmax, "count")


@declare_units(
    tasmin="[temperature]",
    tasmax="[temperature]",
    thresh_tasmin="[temperature]",
    thresh_tasmax="[temperature]",
)
def tx_tn_days_above(
    tasmin: xarray.DataArray,
    tasmax: xarray.DataArray,
    thresh_tasmin: Quantified = "22 degC",
    thresh_tasmax: Quantified = "30 degC",
    freq: str = "YS",
    op: str = ">",
) -> xarray.DataArray:
    r"""Number of days with both hot maximum and minimum daily temperatures.

    The number of days per period with tasmin above a threshold and tasmax above another threshold.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh_tasmin : Quantified
        Threshold temperature for tasmin on which to base evaluation.
    thresh_tasmax : Quantified
        Threshold temperature for tasmax on which to base evaluation.
    freq : str
        Resampling frequency.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [time]
      the number of days with tasmin > thresh_tasmin and tasmax > thresh_tasmax per period.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`, :math:`TN_{ij}`
    the daily minimum temperature at day :math:`i` of period :math:`j`, :math:`TX_{thresh}` the threshold for maximum
    daily temperature, and :math:`TN_{thresh}` the threshold for minimum daily temperature. Then counted is the number
    of days where:

    .. math::

        TX_{ij} > TX_{thresh} [℃]

    and where:

    .. math::

        TN_{ij} > TN_{thresh} [℃]

    """
    thresh_tasmax = convert_units_to(thresh_tasmax, tasmax)
    thresh_tasmin = convert_units_to(thresh_tasmin, tasmin)

    constrain = (">", ">=")
    events = (
        compare(tasmin, op, thresh_tasmin, constrain)
        & compare(tasmax, op, thresh_tasmax, constrain)
    ) * 1
    out = events.resample(time=freq).sum(dim="time")
    return to_agg_units(out, tasmin, "count")


@declare_units(tasmax="[temperature]", tasmax_per="[temperature]")
@percentile_bootstrap
def warm_spell_duration_index(
    tasmax: xarray.DataArray,
    tasmax_per: xarray.DataArray,
    window: int = 6,
    freq: str = "YS",
    resample_before_rl: bool = True,
    bootstrap: bool = False,  # noqa
    op: str = ">",
) -> xarray.DataArray:
    r"""Warm spell duration index.

    Number of days inside spells of a minimum number of consecutive days when the daily maximum temperature is above the
    90th percentile. The 90th percentile should be computed for a 5-day moving window, centered on each calendar day in
    the 1961-1990 period.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    tasmax_per : xarray.DataArray
        percentile(s) of daily maximum temperature.
    window : int
        Minimum number of days with temperature above threshold to qualify as a warm spell.
    freq : str
        Resampling frequency.
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.
    bootstrap : bool
        Flag to run bootstrapping of percentiles. Used by percentile_bootstrap decorator.
        Bootstrapping is only useful when the percentiles are computed on a part of the studied sample.
        This period, common to percentiles and the sample must be bootstrapped to avoid inhomogeneities with
        the rest of the time series.
        Keep bootstrap to False when there is no common period, it would give wrong results
        plus, bootstrapping is computationally expensive.
    op : {">", ">=", "gt", "ge"}
        Comparison operation. Default: ">".

    Returns
    -------
    xarray.DataArray, [time]
        Warm spell duration index.

    References
    ----------
    From the Expert Team on Climate Change Detection, Monitoring and Indices (ETCCDMI; :cite:p:`zhang_indices_2011`).
    Used in :cite:cts:`alexander_global_2006`

    Examples
    --------
    Note that this example does not use a proper 1961-1990 reference period.

    >>> from xclim.core.calendar import percentile_doy
    >>> from xclim.indices import warm_spell_duration_index

    >>> tasmax = xr.open_dataset(path_to_tasmax_file).tasmax.isel(lat=0, lon=0)
    >>> tasmax_per = percentile_doy(tasmax, per=90).sel(percentiles=90)
    >>> warm_spell_duration_index(tasmax, tasmax_per)
    """
    thresh = convert_units_to(tasmax_per, tasmax)

    # Create time series out of doy values.
    thresh = resample_doy(thresh, tasmax)

    above = compare(tasmax, op, thresh, constrain=(">", ">="))
    out = rl.resample_and_rl(
        above,
        resample_before_rl,
        rl.windowed_run_count,
        window=window,
        freq=freq,
    )

    return to_agg_units(out, tasmax, "count")


@declare_units(pr="[precipitation]", prsn="[precipitation]", tas="[temperature]")
def winter_rain_ratio(
    *,
    pr: xarray.DataArray,
    prsn: xarray.DataArray | None = None,
    tas: xarray.DataArray | None = None,
    freq: str = "QS-DEC",
) -> xarray.DataArray:
    """Ratio of rainfall to total precipitation during winter.

    The ratio of total liquid precipitation over the total precipitation over the winter months (DJF). If solid
    precipitation is not provided, then precipitation is assumed solid if the temperature is below 0°C.

    Parameters
    ----------
    pr : xarray.DataArray
        Mean daily precipitation flux.
    prsn : xarray.DataArray, optional
        Mean daily solid precipitation flux.
    tas : xarray.DataArray, optional
        Mean daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        Ratio of rainfall to total precipitation during winter months (DJF).
    """
    ratio = liquid_precip_ratio(pr, prsn, tas, freq=freq)
    winter = ratio.indexes["time"].month == 12
    ratio = ratio.sel(time=winter)
    return ratio


@declare_units(
    snd="[length]", sfcWind="[speed]", snd_thresh="[length]", sfcWind_thresh="[speed]"
)
def blowing_snow(
    snd: xarray.DataArray,
    sfcWind: xarray.DataArray,  # noqa
    snd_thresh: Quantified = "5 cm",
    sfcWind_thresh: Quantified = "15 km/h",  # noqa
    window: int = 3,
    freq: str = "YS-JUL",
) -> xarray.DataArray:
    """Blowing snow days.

    Number of days when both snowfall over the last days and daily wind speeds are above respective thresholds.

    Parameters
    ----------
    snd : xarray.DataArray
        Surface snow depth.
    sfcWind : xr.DataArray
        Wind velocity
    snd_thresh : Quantified
        Threshold on net snowfall accumulation over the last `window` days.
    sfcWind_thresh : Quantified
        Wind speed threshold.
    window : int
        Period over which snow is accumulated before comparing against threshold.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        Number of days when snowfall and wind speeds are above respective thresholds.
    """
    snd_thresh = convert_units_to(snd_thresh, snd)
    sfcWind_thresh = convert_units_to(sfcWind_thresh, sfcWind)  # noqa

    # Net snow accumulation over the last `window` days
    snow = snd.diff(dim="time").rolling(time=window, center=False).sum()

    # Blowing snow conditions
    cond = (snow >= snd_thresh) * (sfcWind >= sfcWind_thresh) * 1

    out = cond.resample(time=freq).sum(dim="time")
    out = out.assign_attrs(units=to_agg_units(out, snd, "count"))
    return out


@declare_units(pr="[precipitation]", evspsbl="[precipitation]")
def water_cycle_intensity(
    pr: xarray.DataArray, evspsbl: xarray.DataArray, freq="YS"
) -> xarray.DataArray:
    """Water cycle intensity.

    The sum of precipitation and actual evapotranspiration.

    Parameters
    ----------
    pr : xarray.DataArray
        Precipitation flux.
    evspsbl : xarray.DataArray
        Actual evapotranspiration flux.

    Returns
    -------
    xarray.DataArray
        The sum of precipitation and actual evapotranspiration for each period.

    References
    ----------
    :cite:cts:`huntington_2018`
    """
    pr = convert_units_to(pr, evspsbl)

    # Water cycle intensity
    wci = (pr + evspsbl).assign_attrs(units=pr.units)
    wci = rate2amount(wci)
    wci = wci.resample(time=freq).sum(dim="time").assign_attrs(units=wci.units)
    return wci



"""
Measures Submodule
==================
SDBA diagnostic tests are made up of properties and measures. Measures compare adjusted simulations to a reference,
through statistical properties or directly. This framework for the diagnostic tests was inspired by the
`VALUE <http://www.value-cost.eu/>`_ project.

"""

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import xarray as xr

from xclim.core.indicator import Indicator, base_registry
from xclim.core.units import convert_units_to, pint2cfattrs, units2pint
from xclim.core.utils import InputKind

from .base import Grouper
from .utils import _pairwise_spearman


class StatisticalMeasure(Indicator):
    """Base indicator class for statistical measures used when validating bias-adjusted outputs.

    Statistical measures use input data where the time dimension was reduced, usually by the computation
    of a :py:class:`xclim.sdba.properties.StatisticalProperty` instance.  They usually take two arrays
    as input: "sim" and "ref", "sim" being measured against "ref". The two arrays must have identical
    coordinates on their common dimensions.

    Statistical measures are generally unit-generic. If the inputs have different units, "sim" is converted
    to match "ref".
    """

    realm = "generic"

    @classmethod
    def _ensure_correct_parameters(cls, parameters):
        inputs = {k for k, p in parameters.items() if p.kind == InputKind.VARIABLE}
        if not inputs.issuperset({"sim", "ref"}):
            raise ValueError(
                f"{cls.__name__} requires 'sim' and 'ref' as inputs. Got {inputs}."
            )
        return super()._ensure_correct_parameters(parameters)

    def _preprocess_and_checks(self, das, params):
        """Perform parent's checks and also check convert units so that sim matches ref."""
        das, params = super()._preprocess_and_checks(das, params)

        # Convert grouping and check if allowed:
        sim = das["sim"]
        ref = das["ref"]

        # Check if common coordinates are identical.
        newsim, newref = xr.broadcast(sim, ref)
        for dim in set(sim.dims).union(ref.dims):
            if [sim[dim].size, ref[dim].size] != [newsim[dim].size, newref[dim].size]:
                raise ValueError(
                    f"Common dimension {dim} has different coordinates between ref and sim."
                )

        units_sim = units2pint(sim)
        units_ref = units2pint(ref)

        if units_sim != units_ref:
            das["sim"] = convert_units_to(sim, ref)

        return das, params


class StatisticalPropertyMeasure(Indicator):
    """Base indicator class for statistical properties that include the comparison measure, used when validating bias-adjusted outputs.

    StatisticalPropertyMeasure objects combine the functionalities of
    :py:class:`xclim.sdba.properties.StatisticalProperty` and
    :py:class:`xclim.sdba.properties.StatisticalMeasure`.

    Statistical properties usually reduce the time dimension and sometimes more dimensions
    (for example in spatial properties), sometimes adding a grouping dimension according to
    the passed value of `group` (e.g.: group='time.month' means the loss of the time dimension
    and the addition of a month one).

    Statistical measures usually take two arrays as input: "sim" and "ref", "sim" being measured against "ref".

    Statistical property-measures are generally unit-generic. If the inputs have different units,
    "sim" is converted to match "ref".
    """

    aspect = None
    """The aspect the statistical property studies: marginal, temporal, multivariate or spatial."""

    allowed_groups = None
    """A list of allowed groupings. A subset of dayofyear, week, month, season or group.
    The latter stands for no temporal grouping."""

    realm = "generic"

    @classmethod
    def _ensure_correct_parameters(cls, parameters):
        inputs = {k for k, p in parameters.items() if p.kind == InputKind.VARIABLE}
        if not inputs.issuperset({"sim", "ref"}):
            raise ValueError(
                f"{cls.__name__} requires 'sim' and 'ref' as inputs. Got {inputs}."
            )

        if "group" not in parameters:
            raise ValueError(
                f"{cls.__name__} require a 'group' argument, use the base Indicator"
                " class if your computation doesn't perform any regrouping."
            )

        return super()._ensure_correct_parameters(parameters)

    def _preprocess_and_checks(self, das, params):
        """Perform parent's checks and also check convert units so that sim matches ref."""
        das, params = super()._preprocess_and_checks(das, params)

        # Convert grouping and check if allowed:
        if isinstance(params["group"], str):
            params["group"] = Grouper(params["group"])

        if self.allowed_groups is not None:
            if params["group"].prop not in self.allowed_groups:
                raise ValueError(
                    f"Grouping period {params['group'].prop_name} is not allowed for property "
                    f"{self.identifier} (needs something in "
                    f"{list(map(lambda g: '<dim>.' + g.replace('group', ''), self.allowed_groups))})."
                )

        # Convert grouping and check if allowed:
        sim = das["sim"]
        ref = das["ref"]
        units_sim = units2pint(sim)
        units_ref = units2pint(ref)

        if units_sim != units_ref:
            das["sim"] = convert_units_to(sim, ref)

        return das, params

    def _postprocess(self, outs, das, params):
        """Squeeze `group` dim if needed."""
        outs = super()._postprocess(outs, das, params)

        for ii, out in enumerate(outs):
            if "group" in out.dims:
                outs[ii] = out.squeeze("group", drop=True)

        return outs


base_registry["StatisticalMeasure"] = StatisticalMeasure
base_registry["StatisticalPropertyMeasure"] = StatisticalPropertyMeasure


def _bias(sim: xr.DataArray, ref: xr.DataArray) -> xr.DataArray:
    """Bias.

    The bias is the simulation minus the reference.

    Parameters
    ----------
    sim : xr.DataArray
      data from the simulation (one value for each grid-point)
    ref : xr.DataArray
      data from the reference (observations) (one value for each grid-point)

    Returns
    -------
    xr.DataArray, [same as ref]
      Absolute bias
    """
    out = sim - ref
    out.attrs.update(pint2cfattrs(units2pint(ref.attrs["units"]), is_difference=True))
    return out


bias = StatisticalMeasure(identifier="bias", compute=_bias)


def _relative_bias(sim: xr.DataArray, ref: xr.DataArray) -> xr.DataArray:
    """Relative Bias.

    The relative bias is the simulation minus reference, divided by the reference.

    Parameters
    ----------
    sim : xr.DataArray
      data from the simulation (one value for each grid-point)
    ref : xr.DataArray
      data from the reference (observations) (one value for each grid-point)

    Returns
    -------
    xr.DataArray, [dimensionless]
      Relative bias
    """
    out = (sim - ref) / ref
    return out.assign_attrs(units="")


relative_bias = StatisticalMeasure(
    identifier="relative_bias", compute=_relative_bias, units=""
)


def _circular_bias(sim: xr.DataArray, ref: xr.DataArray) -> xr.DataArray:
    """Circular bias.

    Bias considering circular time series.
    E.g. The bias between doy 365 and doy 1 is 364, but the circular bias is -1.

    Parameters
    ----------
    sim : xr.DataArray
      data from the simulation (one value for each grid-point)
    ref : xr.DataArray
      data from the reference (observations) (one value for each grid-point)

    Returns
    -------
    xr.DataArray, [days]
      Circular bias
    """
    out = (sim - ref) % 365
    out = out.where(
        out <= 365 / 2, 365 - out
    )  # when condition false, replace by 2nd arg
    out = out.where(ref >= sim, out * -1)  # when condition false, replace by 2nd arg
    return out.assign_attrs(units="days")


circular_bias = StatisticalMeasure(
    identifier="circular_bias", compute=_circular_bias, units="days"
)


def _ratio(sim: xr.DataArray, ref: xr.DataArray) -> xr.DataArray:
    """Ratio.

    The ratio is the quotient of the simulation over the reference.

    Parameters
    ----------
    sim : xr.DataArray
      data from the simulation (one value for each grid-point)
    ref : xr.DataArray
      data from the reference (observations) (one value for each grid-point)

    Returns
    -------
    xr.DataArray, [dimensionless]
      Ratio
    """
    out = sim / ref
    out.attrs["units"] = ""
    return out


ratio = StatisticalMeasure(identifier="ratio", compute=_ratio, units="")


def _rmse(
    sim: xr.DataArray, ref: xr.DataArray, group: str | Grouper = "time"
) -> xr.DataArray:
    """Root mean square error.

    The root mean square error on the time dimension between the simulation and the reference.

    Parameters
    ----------
    sim : xr.DataArray
        Data from the simulation (a time-series for each grid-point)
    ref : xr.DataArray
        Data from the reference (observations) (a time-series for each grid-point)
    group: str
        Compute the property and measure for each temporal groups individually.
        Currently not implemented.

    Returns
    -------
    xr.DataArray, [same as ref]
      Root mean square error
    """

    def _rmse_internal(_sim: xr.DataArray, _ref: xr.DataArray) -> xr.DataArray:
        _f: xr.DataArray = np.sqrt(np.mean((_sim - _ref) ** 2, axis=-1))
        return _f

    out = xr.apply_ufunc(
        _rmse_internal,
        sim,
        ref,
        input_core_dims=[["time"], ["time"]],
        dask="parallelized",
    )
    out = out.assign_attrs(pint2cfattrs(units2pint(ref.units), is_difference=True))
    return out


rmse = StatisticalPropertyMeasure(
    identifier="rmse",
    aspect="temporal",
    compute=_rmse,
    allowed_groups=["group"],
    cell_methods="time: mean",
)


def _mae(
    sim: xr.DataArray, ref: xr.DataArray, group: str | Grouper = "time"
) -> xr.DataArray:
    """Mean absolute error.

    The mean absolute error on the time dimension between the simulation and the reference.

    Parameters
    ----------
    sim : xr.DataArray
        data from the simulation (a time-series for each grid-point)
    ref : xr.DataArray
        data from the reference (observations) (a time-series for each grid-point)
    group : str
        Compute the property and measure for each temporal groups individually.
        Currently not implemented.

    Returns
    -------
    xr.DataArray, [same as ref]
      Mean absolute error
    """

    def _mae_internal(_sim: xr.DataArray, _ref: xr.DataArray) -> xr.DataArray:
        _f: xr.DataArray = np.mean(np.abs(_sim - _ref), axis=-1)
        return _f

    out = xr.apply_ufunc(
        _mae_internal,
        sim,
        ref,
        input_core_dims=[["time"], ["time"]],
        dask="parallelized",
    )
    out = out.assign_attrs(pint2cfattrs(units2pint(ref.units), is_difference=True))
    return out


mae = StatisticalPropertyMeasure(
    identifier="mae",
    aspect="temporal",
    compute=_mae,
    allowed_groups=["group"],
    cell_methods="time: mean",
)


def _annual_cycle_correlation(
    sim: xr.DataArray,
    ref: xr.DataArray,
    window: int = 15,
    group: str | Grouper = "time",
) -> xr.DataArray:
    """Annual cycle correlation.

    Pearson correlation coefficient between the smooth day-of-year averaged annual cycles of the simulation and
    the reference. In the smooth day-of-year averaged annual cycles, each day-of-year is averaged over all years
    and over a window of days around that day.

    Parameters
    ----------
    sim : xr.DataArray
        data from the simulation (a time-series for each grid-point)
    ref : xr.DataArray
        data from the reference (observations) (a time-series for each grid-point)
    window : int
        Size of window around each day of year around which to take the mean.
        E.g. If window=31, Jan 1st is averaged over from December 17th to January 16th.
    group : str
        Compute the property and measure for each temporal groups individually.
        Currently not implemented.

    Returns
    -------
    xr.DataArray, [dimensionless]
      Annual cycle correlation
    """
    # group by day-of-year and window around each doy
    grouper_test = Grouper("time.dayofyear", window=window)
    # for each day, mean over X day window and over all years to create a smooth avg annual cycle
    sim_annual_cycle = grouper_test.apply("mean", sim)
    ref_annual_cycle = grouper_test.apply("mean", ref)
    out = xr.corr(ref_annual_cycle, sim_annual_cycle, dim="dayofyear")
    return out.assign_attrs(units="")


annual_cycle_correlation = StatisticalPropertyMeasure(
    identifier="annual_cycle_correlation",
    aspect="temporal",
    compute=_annual_cycle_correlation,
    allowed_groups=["group"],
)


def _scorr(
    sim: xr.DataArray,
    ref: xr.DataArray,
    *,
    dims: Sequence | None = None,
    group: str | Grouper = "time",
):
    """Spatial correllogram.

    Compute the inter-site correlations of each array, compute the difference in correlations and sum.
    Taken from Vrac (2018). The spatial and temporal dimensions are reduced.

    Parameters
    ----------
    sim : xr.DataArray
        data from the simulation (a time-series for each grid-point)
    ref : xr.DataArray
        data from the reference (observations) (a time-series for each grid-point)
    dims : sequence of strings, optional
        Name of the spatial dimensions. If None (default), all dimensions except 'time' are used.
    group : str
        Compute the property and measure for each temporal groups individually.
        Currently not implemented.

    Returns
    -------
    xr.DataArray, [dimensionless]
      Sum of the inter-site correlation differences.
    """
    if dims is None:
        dims = [d for d in sim.dims if d != "time"]

    refcorr = _pairwise_spearman(ref, dims)
    simcorr = _pairwise_spearman(sim, dims)
    S_corr = (simcorr - refcorr).sum(["_spatial", "_spatial2"])
    return S_corr.assign_attrs(units="")


scorr = StatisticalPropertyMeasure(
    identifier="Scorr", aspect="spatial", compute=_scorr, allowed_groups=["group"]
)


def _taylordiagram(
    sim: xr.DataArray,
    ref: xr.DataArray,
    dim: str = "time",
    group: str | Grouper = "time",
    normalize: bool = False,
) -> xr.DataArray:
    """Taylor diagram.

    Compute the respective standard deviations of a simulation and a reference array, as well as the Pearson
    correlation coefficient between both, all necessary parameters to plot points on a Taylor diagram.

    Parameters
    ----------
    sim : xr.DataArray
        data from the simulation (a time-series for each grid-point)
    ref : xr.DataArray
        data from the reference (observations) (a time-series for each grid-point)
    dim : str
        Dimension across which the correlation and standard deviation should be computed.
    group : str
        Compute the property and measure for each temporal groups individually.
        Currently not implemented.
    normalize : bool
        If `True`, divide the standard deviations by the standard deviation of the reference.
        Default is `False`.


    Returns
    -------
    xr.DataArray, [same as ref]
        Standard deviations of sim, ref and correlation coefficient between both.
    """
    corr = xr.corr(sim, ref, dim=dim)

    ref_std = ref.std(dim=dim, skipna=True, keep_attrs=True)
    sim_std = sim.std(dim=dim, skipna=True, keep_attrs=True)

    new_dim = xr.DataArray(
        ["ref_std", "sim_std", "corr"], dims=("taylor_param",), name="taylor_param"
    )

    out = xr.concat(
        [ref_std, sim_std, corr],
        new_dim,
        coords="minimal",
        compat="override",  # Take common coords from `ref_std`.
    ).assign_attrs(
        {
            "correlation_type": "Pearson correlation coefficient",
            "units": ref.units,
        }
    )

    # Normalize the standard deviations byt the standard deviation of the reference.
    if normalize:
        if (out[{"taylor_param": 0}] == 0).any():
            raise ValueError(
                "`ref_std =0` (homogeneous field) obtained, normalization is not possible."
            )
        with xr.set_options(keep_attrs=True):
            out[{"taylor_param": [0, 1]}] = (
                out[{"taylor_param": [0, 1]}] / out[{"taylor_param": 0}]
            )
        out.attrs["normalized"] = True
        out.attrs["units"] = ""

    return out


taylordiagram = StatisticalPropertyMeasure(
    identifier="taylordiagram",
    aspect="temporal",
    compute=_taylordiagram,
    allowed_groups=["group"],
)
